package de.upb.infexba.ae.categorization;

import java.io.FileInputStream;
import java.io.ObjectInputStream;
import java.util.ArrayList;
import java.util.List;

import org.apache.uima.UimaContext;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;
import org.apache.uima.resource.ResourceInitializationException;

import weka.classifiers.Classifier;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.SparseInstance;
import weka.core.converters.ConverterUtils.DataSource;
import de.upb.infexba.Genre;
import de.upb.infexba.Sentence;
import de.upb.infexba.ae.InfexBAAnalysisEngine;
import de.upb.infexba.ml.GeneralFeatureARFFReader;
import de.upb.infexba.ml.IntentionFeatureGenerator;


public class IntentionAnalyzer extends InfexBAAnalysisEngine{
	
	/**
	 * Config parameter for setting the location of the classifier model file
	 */
	public static final String PARAM_MODEL_FILE = "ModelFile";

	/**
	 * Config parameter for setting the location of the classifier dataset file
	 */
	public static final String PARAM_DATASET_FILE = "DatasetFile";
	
	
	// -------------------------------------------------------------------
	// VALUES AND OBJECTS
	// -------------------------------------------------------------------
	
	/**
	 * Indicates whether the classifier works on document level, i.e. it 
	 * classifies the given document as a whole
	 */
	private static final boolean DOCUMENT_LEVEL_CLASSIFICATION = true;
	
	/**
	 * Indicates whether the classifier works on sentence level, i.e. it 
	 * classifies each sentence of the given document
	 */
	private static final boolean SENTENCE_LEVEL_CLASSIFICATION = false;
	
	
	/**
	 * The classifier to be used for classification. You might need two further
	 * classifier if you solve the task as a triple binary classification task.
	 */
	private static Classifier classifier;
	
	/**
	 * The dataset description to be used by the classifier of the machine 
	 * learning approach
	 */
	private static Instances dataSet;
	
	/**
	 * The feature generator 
	 */
	private IntentionFeatureGenerator generator;
	
	
	// -------------------------------------------------------------------
	// INITIALIZATION
	// -------------------------------------------------------------------
	
	@Override
	public void initialize(UimaContext aContext) throws ResourceInitializationException {
		super.initialize(aContext);

		try {
			if (super.isActivated()){				
				// Initialize classifier
				String modelLocation = (String) aContext.getConfigParameterValue(PARAM_MODEL_FILE);
				ObjectInputStream ois = new ObjectInputStream(
						new FileInputStream(modelLocation));
				classifier = (Classifier) ois.readObject();
				// Initialize dataset
				String datasetLocation = (String) aContext.getConfigParameterValue(PARAM_DATASET_FILE);
				DataSource source = new DataSource(datasetLocation);
				dataSet = source.getDataSet();
				dataSet.setClassIndex(dataSet.numAttributes() - 1);
				
				// Initialize feature generation
				generator = new IntentionFeatureGenerator();
				GeneralFeatureARFFReader reader = new GeneralFeatureARFFReader();
				List<String> featureNames = reader.readFeatureSetFromARFF(datasetLocation);
				generator.setFeatureSet(featureNames);
			}
		}
		catch (Exception e){
			e.printStackTrace();
		}
	}

	
	
	// -------------------------------------------------------------------
	// PROCESSING
	// -------------------------------------------------------------------
	
	@Override
	public void processDocument(JCas jcas){
		String classification = null;
		
		if (DOCUMENT_LEVEL_CLASSIFICATION){
			// Only classify the whole document
			classification = this.classify(jcas, new Annotation(jcas, 0, jcas.getDocumentText().length()));
		}
		
		else if (SENTENCE_LEVEL_CLASSIFICATION){
			// Classify each sentence of the document
			List<String> sentenceIntentions = new ArrayList<String>();
			FSIterator<Annotation> sentenceIter = jcas.getAnnotationIndex(Sentence.type).iterator();
			while (sentenceIter.hasNext()){
				Sentence sentence = (Sentence) sentenceIter.next();
				sentenceIntentions.add(this.classify(jcas, sentence));
			}
			// Now the overall classification must be derived
			classification = this.inferDocumentClassification(jcas, sentenceIntentions);
		}
		
		// Create new intention annotation
		this.createAnnotation(jcas, classification);
	}

	private String classify(JCas jcas, Annotation annotation){	
		// Create instance
		Instance instance = new SparseInstance(1, generator.computeFeatures(jcas, annotation));
		instance.setDataset(dataSet);
		try{
			return this.classifyInstance(instance);	
		}
		catch (Exception ex){
			ex.printStackTrace();
		}	
		return null;
	}
	
	
	private String inferDocumentClassification(JCas jcas, List<String> sentenceAnnotations){
		// TODO Needs to be changed according to what to do with the sentence annotations
		return null;
	}
	
	private String classifyInstance(Instance instance) throws Exception {
		// TODO Needs to be changed to make your classifier do the right thing
		if (classifier.classifyInstance(instance) == 1)
			return null;
		else return null;
	}
	
	
	private void createAnnotation(JCas jcas, String classification){
		Genre intention = new Genre(jcas, 0, jcas.getDocumentText().length());
		intention.setGenreClass(classification);
		intention.addToIndexes();
	}
}


