package de.upb.infexba.ae.preprocessing;

import is2.data.SentenceData09;
import is2.tag5.Options;
import is2.tag5.Tagger;

import java.util.List;

import org.apache.uima.UimaContext;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;
import org.apache.uima.resource.ResourceInitializationException;

import de.upb.infexba.Sentence;
import de.upb.infexba.Token;
import de.upb.infexba.ae.InfexBAAnalysisEngine;
import de.upb.infexba.util.AnnoTools;

/**
 * Analysis engine that wraps the part-of-speech (POS) tagger included in the 
 * <A HREF="http://code.google.com/p/mate-tools/" target="_blank">mate-tools</A> of 
 * Bernd Bohnet, University of Stuttgart. 
 * Given token and sentence annotations, this analysis engine computes the part-of-speech
 * tag for each token.
 * It is language independent that means no lexicon or morphology is used which would
 * provide extra data. Everything is aquired during the training and no manual langauge 
 * dependent adaption is needed.<BR><BR>
 * 
 * The POS tagger is based on a statistical model. The location of this model has
 * to be set via the parameter <CODE>MateToolsPOSTaggerModel</CODE>.
 * 
 * @author hwachsmuth, 10/15/2010
 *
 */
public class MateToolsPOSTaggerWrapper extends InfexBAAnalysisEngine{
	
	// -----------------------------------------------------------------------------------
	// INITIALIZATION
	// -----------------------------------------------------------------------------------
	
	/**
	 * Config parameter for the model used by the POS tagger 
	 */
	public static final String PARAM_MODEL = "MateToolsPOSTaggerModel";
	
	/**
	 * The mate-tools POS tagger
	 */
	private Tagger posTagger;
	
	/**
	 * Options for the POS tagger
	 */
	private Options optsTagger;
	
	@Override
	public void initialize(UimaContext aContext) throws ResourceInitializationException {
		super.initialize(aContext);
		// Create the POS tagger with a model set via parameter
		String model = (String) aContext.getConfigParameterValue(PARAM_MODEL);
		if (super.isActivated()){
			optsTagger = new Options(new String[] {"-model", model});
			posTagger = new Tagger(optsTagger);
		}
	}

	
	// -----------------------------------------------------------------------------------
	// PROCESSING
	// -----------------------------------------------------------------------------------
	
	@Override
	public void processDocument(JCas jcas){
		this.tagDocument(jcas);
	}
	
	/**
	 * POS tag all tokens of the document represented in the JCas object.
	 * 
	 * @param jcas The JCas object
	 */
	private void tagDocument(JCas jcas){
		FSIterator<Annotation> sentenceIter = jcas.getAnnotationIndex(Sentence.type).iterator();
		while (sentenceIter.hasNext())
			this.tagSentence(jcas, (Sentence) sentenceIter.next());
	}

	/**
	 * POS tag all tokens of the given sentence.
	 * 
	 * @param jcas The JCas object
	 * @param sentence The sentence
	 */
	private void tagSentence(JCas jcas, Sentence sentence){
		// Get tokens and convert to mate-tools structure
		List<Token> tokens = AnnoTools.sentenceToTokens(jcas, sentence);
		List<String> tokenStrings = AnnoTools.tokensToStrings(tokens);
		SentenceData09 sentenceData = new SentenceData09();
		sentenceData.init(tokenStrings.toArray(new String[0]));
		// Tag and relate POS tags to tokens
		posTagger.tag(sentenceData);
		for(int i=0; i<sentenceData.ppos.length; i++)
			tokens.get(i).setPos(sentenceData.ppos[i]);
	}
}
