package de.upb.infexba.ae.preprocessing;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;

import de.upb.infexba.Sentence;
import de.upb.infexba.ae.InfexBAAnalysisEngine;
import de.upb.infexba.linguistic.RegexTime;
import de.upb.infexba.util.CharTools;
import de.upb.infexba.util.WordLists;

/**
 * <p>
 * This analysis engine represents a basic rule-based sentence splitter 
 * which, in general, can be used to split any German text into sentences. 
 * However, the rules used by the splitter are specified with German news 
 * articles given as plain text in mind. 
 * <p>
 * Namely, the essential currently used rules for classyfing whether to 
 * split the sentences at the current text position are:
 * <ul>
 *   <li><b>IF</b> a question mark or an exclamation mark is followed by a 
 *       space<br> 
 *       <b>THEN</b> the sentence ends after the corresponding mark.</li><br>
 *   <li><b>IF</b> a question mark or exclamation mark or period is followed by 
 *       a quotation mark<br> 
 *       <b>THEN</b> the sentence ends after or before that mark depending on 
 *         whether there was a quotation mark at the sentence begin.</li><br>
 *       
 *       
 *   <li><b>IF</b> a period is followed by a space<br> 
 *       <b>THEN</b> the sentence ends after the period if<br>
 *       <ul>
 *         <li>a whitespace is following, <b>AND</b></li>
 *         <li>the next letter is not upper case, <b>AND</b></li>
 *         <li>the preceding character is not a single one, <b>AND</b></li>
 *         <li>the preceding token doesn't seem to be ordinal, <b>AND</b></li>
 *         <li>it does not belong to a known acronym. <b>OR</b> if</li>
 *         <li>it a typical sentence start word is following.</li>
 *       </ul></li>
 *       
 *   <li><b>IF</b> a line only consists of upper case letters<br> 
 *       <b>THEN</b> the corresponding sentence ends at the end of the 
 *       line.</li><br>
 *   <li><b>IF</b> one or more lines only consists of one subsequent URL<br> 
 *       <b>THEN</b> the corresponding sentence ends at the end of these 
 *       lines.</li><br>
 *   <li><b>IF</b> one or more lines are completely bracketed<br> 
 *       <b>THEN</b> the sentence ends at the end of these lines.</li><br>
 *       
 *   <li><b>IF</b> a blank line follows next<br>
 *       <b>THEN</b> the sentence ends after the current character.</li><br>
 *   <li><b>IF</b> this sentence refers to a bulletpoint and another 
 *       bulletpoint follows next<br>
 *       <b>THEN</b> the sentence ends after the current character.</b></li><br>
 *    
 *   <li><b>IF</b> there is a sequence of more than one hyphen<br> 
 *       <b>THEN</b> the sentence ends after these hyphens.</li>
 * </ul>
 * <p>
 * (In all rules, it is implied that the end of the previous sentence is
 *  known correctly)
 * <p>
 * This analysis engine is not dependent from any other analysis engine, in
 * particular, it does not need a tokenized text to work.
 * <p>
 * The sentence splitter achieved about 97% to 98% recall in evaluation
 * on the revenue corpus!!! Precision couldn't be measured but since every 
 * sentence is annotated, it can be assumed to be accordingly high.
 * 
 * @author hwachsmuth
 *
 */
public class RBSentenceSplitter extends InfexBAAnalysisEngine{
    
    
		
	// -------------------------------------------------------------------
	// CONSTANTS
	// -------------------------------------------------------------------
	
	/**
	 * Lower bound of range in which numbers are assumed to be possible
	 * ordinal numbers.
	 */
	private static final int RANGE_LOWER_BOUND = 1;
	
	/**
	 * Upper bound of range in which numbers are assumed to be possible
	 * ordinal numbers.
	 */
	private static final int RANGE_UPPER_BOUND = 100;
	
	/**
	 * Start substring of a string to be assumed to be a URL
	 */
	private static final String URL_INDICATOR = "http://";
	
	/**
	 * The character assumed to be before the first character of a
	 * document.
	 */
	private static final char TEXT_START_CHARACTER = ' ';

	/**
	 * The character assumed to be after the last character of a
	 * document.
	 */
	private static final char TEXT_END_CHARACTER = '\n';
	
	/**
	 * Minimum number of characters a line must have to be assumed to be
	 * continued in the next line
	 */
	private static final int REGULAR_LINE_MIN_LENGTH = 50;
	
	
	// -------------------------------------------------------------------
	// PROCESSING A TEXT
	// -------------------------------------------------------------------
	
	@Override
	public void processDocument(JCas jcas){
		split(jcas);
	}
	
	/**
	 * Splits the doument text given in the <code>JCas</code> object into its
	 * contained sentences. That is, it creates a new annotation for every
	 * found sentence and adds it to the indices.
	 * 
	 * @param jcas The JCas object
	 */
	private void split(JCas jcas){
		String text = jcas.getDocumentText();
		int index = 0;
		while (index < text.length()-1){
			Annotation sentence = findNextSentence(jcas, text, index);
			sentence.addToIndexes();
			// Skip every character between two sentences that is a space or hyphen
			index = sentence.getEnd()+1;
			while (index < text.length()-1 
					&& !Character.isLetterOrDigit(text.charAt(index)) 
					&& !CharTools.isCurrencySymbol(text.charAt(index)))
				index++;
			// Don't wanna miss specific symbols at beginning of sentence
			if (characterMayBeSentenceBegin(text, index-1))
				index--;
		}
	}
	
	/**
	 * Determines the next sentence starting at the index after the last sentence 
	 * (i.e. <code>begin</code>). This method based on several heuristic rules. 
	 * Notice that most rules use the notion that a sentence end just ended implictly.
	 *  
	 * @param jcas The JCas object
	 * @param text The text to find the next sentence in
	 * @param begin The start index of the sentence to be found
	 * @return A sentence annotation
	 */
	private Annotation findNextSentence(JCas jcas, String text, int begin){
		// Helper for special cases: upper case, url, and bracketed sentences
		boolean isUpperCase = Character.isUpperCase(text.charAt(begin));
		boolean isURL = isHeadOfURL(text, begin);
		boolean isInBrackets = CharTools.isOpeningBracket(text.charAt(begin));
		boolean isQuotation = CharTools.isQuotationMark(text.charAt(begin));
		
		// Helper for special case: Bullet lists and title lines
		int lineCount = begin-this.getIndexOfLineBegin(text, begin);
	    int lineBegin = begin-lineCount;
	    while (lineBegin < text.length()-1 && Character.isWhitespace(text.charAt(lineBegin)))
	    	lineBegin++;
		char firstChar = text.charAt(lineBegin);
		boolean isList = CharTools.isBullet(firstChar); 
		
		// Search for end index of current sentence
		for (int end = begin; end<text.length(); end++){
			lineCount++;
			char cur = text.charAt(end); 
			char succ = TEXT_END_CHARACTER;
			if (end<text.length()-1) succ = text.charAt(end+1);
			char prev = TEXT_START_CHARACTER;
			if (end>0) prev = text.charAt(end-1);
			
			// Maybe split because of following space
			if (Character.isWhitespace(succ)){		
				// If sentence was a URL until here, URL ends now
				isURL = false;
				
				// Unambiguous splitter term...
				if (CharTools.isUmambiguousEndMark(cur))
						return new Sentence(jcas, begin, end+1);
				
				// ... or end of quote at end of sentence
				if (CharTools.isQuotationMark(cur) && CharTools.isFullStop(prev))
					return new Sentence(jcas, begin, end+1);
				
				// Period is splitter term if it doesn't refer to preceding term...
				if (CharTools.isPeriod(cur) && !periodBelongsToPrecedingTerm(text, end)) {
					if (isQuotation) begin+=1;
					return new Sentence(jcas, begin, end+1);
				}
			}
			
			// Maybe split because of period
			if (CharTools.isPeriod(cur)){			
				// If next word is sentence begin...
				if (this.nextWordMayBeSentenceBegin(text, end+1))
					return new Sentence(jcas, begin, end+1);	
				// ... or if there is a footnote after the period
				else if (this.nextIsFootNote(text, end+1)){
					while (end<text.length()-1 && !Character.isWhitespace(text.charAt(end)))
						end++;
					return new Sentence(jcas, begin, end);
				}
			}
			
			// Maybe split because of following line break
			if (Character.isWhitespace(succ) && !Character.isSpaceChar(succ)){
				// This sentence is an upper case sentence...
				if ((lineCount < REGULAR_LINE_MIN_LENGTH && isUpperCase)		
						// ... or it is a single URL
						|| isURL 
						// ... or it is completely bracketed
						|| (isInBrackets && CharTools.isClosingBracket(cur))
						// ... or it is a title line...	
						|| (lineCount < REGULAR_LINE_MIN_LENGTH && precedingTermIsTimeOrDate(text, begin, end))				
						|| (lineCount < REGULAR_LINE_MIN_LENGTH && nextLineStartsWithBullet(text, end+1))
//						|| (lineCount < REGULAR_LINE_MIN_LENGTH && previousLineIsBlankLine(text, begin))	
						// ... or next line is another bulletpoint...
						|| (isList && nextLineStartsWithSameCharacter(text, firstChar, end))
						// ... or next line is a blank line
						|| nextLineIsBlankLine(text, end)
						// ... or next word is sentence begin
						|| nextWordMayBeSentenceBegin(text, end+1)
						){	
					// Go back to last letter (etc.) and split
					while (end > 0 && Character.isWhitespace(text.charAt(end)))
						end -=1;
					return new Sentence(jcas, begin, end+1);
				}
				lineCount = 0;
			}
			
			if (CharTools.isHyphen(cur)){
				// Sentence ends before a sequence of hyphens
				if (CharTools.isHyphen(prev) && !CharTools.isClosingBracket(succ))
					return new Sentence(jcas, begin, end-1);
				// Some sentences end before a separating hyphen
				if (lineBegin == begin && lineCount < REGULAR_LINE_MIN_LENGTH && 
						Character.isSpaceChar(succ)){
					// Skip whitespace
					int i = end+2;
					while (!Character.isLetter(text.charAt(i))) i++;
					// Only sentence end if next word is sentence begin
					if (nextWordMayBeSentenceBegin(text, i)){
						while (end > 0 && Character.isSpaceChar(text.charAt(end-1)))
							end--;
						return new Sentence(jcas, begin, end);
					}
				}
			}
			
			// Set helpers
			if (Character.isLowerCase(cur)) isUpperCase = false;
			if (CharTools.isClosingBracket(cur)) isInBrackets = false;
			if (isQuotation && begin != end && CharTools.isQuotationMark(cur)) 
				isQuotation = false;
		}
		
		// If no end was found before, take the rest
		return new Sentence(jcas, begin, text.length());		
	}
	

	
	// -------------------------------------------------------------------
	// GETTERS
	// -------------------------------------------------------------------
	
	/**
	 * Getter for the index of the first character in the current line.
	 * 
	 * @param text The textt to be analysed
	 * @param i The index of the current character
	 * @return The index of the first character in the current line
	 */
	private int getIndexOfLineBegin(String text, int i){
		int lineBegin = i;
		while (lineBegin>0
				&& !(Character.isWhitespace(text.charAt(lineBegin)) 
						&& !Character.isSpaceChar(text.charAt(lineBegin)))){
				lineBegin--;
		}
		while (lineBegin<text.length()-1 && Character.isWhitespace(text.charAt(lineBegin)))
				lineBegin++;
		return lineBegin;
	}
	
	
	
	// -------------------------------------------------------------------
	// DETERMINERS FOR SPECIAL TERMS
	// -------------------------------------------------------------------
	
	private boolean precedingTermIsTimeOrDate(String text, int begin, int end){
		while (end > 0 && Character.isWhitespace(text.charAt(end))){
			end -=1;
		}
		end+=1;
		
		Pattern datePattern = Pattern.compile(RegexTime.DATE_EXPRESSION);
		Matcher dateMatcher = datePattern.matcher(text.substring(begin, end));
		while (dateMatcher.find())
			if (dateMatcher.end()==end) return true;
		
		Pattern timePattern = Pattern.compile("\\d\\d?:\\d\\d(\\s?Uhr|h)?");
		Matcher timeMatcher = timePattern.matcher(text.substring(begin, end));
		while (timeMatcher.find())
			if (timeMatcher.end()==end) return true;
		
		return false;
	}
	
	/**
	 * Determines whether the successing characters of the one given at
	 * index <code>i</code> could be the beginning characters of a URL.
	 * 
	 * @param text The textt to be analysed
	 * @param i The index of the current character
	 * @return <code>true</code> iff. the defined characters could be the
	 * beginning of a URL
	 */
	private boolean isHeadOfURL(String text, int i){
		if (text.length() - i > URL_INDICATOR.length())
			return text.substring(
					i, i+URL_INDICATOR.length()).toLowerCase().equals(
							URL_INDICATOR);
		return false;
	}
	
	
	
	// -------------------------------------------------------------------
	// DETERMINERS FOR PERIOD REQUESTS
	// -------------------------------------------------------------------
	
	/**
	 * Determines whether the the character at <code>i</code> (which should 
	 * always be a period) belongs to the term that includes at least the 
	 * character at i-1.
	 * 
	 * @param text The text to be analysed
	 * @param i The index of the period
	 * @return <code>true</code> iff. period belongs to the preceding token
	 */
	private boolean periodBelongsToPrecedingTerm(String text, int i){
		if (periodBelongsToSingleCharacter(text, i))
			return true;	
		if (periodBelongsToAnOrdinalNumber(text, i))
			return true;
		if (periodBelongsToAnAcronym(text, i))
			return true;
//		if (! successingTermIsSentenceBegin(text, i))
//			return true;
		return false;		
	}

	/**
	 * Determines whether the period at index <code>i</code> belongs to a
	 * single character, i.e. neither a letter nor a number precedes that character.
	 * 
	 * @param text The text to be analysed
	 * @param i The index of the current character
	 * @return <code>true</code> iff. preceding term is a single character
	 */
	private boolean periodBelongsToSingleCharacter(String text, int i){
		return Character.isLetter(text.charAt(i-1)) 
				&& !Character.isLetterOrDigit(text.charAt(i-2));
	}
	
	/**
	 * Determines whether the period at index <code>j</code> belongs to an ordinal 
	 * number in the range from <code>RANGE_LOWER_BOUND</code> to 
	 * <code>RANGE_UPPER_BOUND</code>. In this case, the number is assumed to be 
	 * an ordinal number.
	 * 
	 * @param text The text to be analysed
	 * @param j The index of the period
	 * @return <code>true</code> iff. preceding term is a number in the demanded 
	 * range
	 */
	private boolean periodBelongsToAnOrdinalNumber(String text, int j){
		int i = j-1;
		if (!Character.isDigit(text.charAt(i))) return false;
		while (i>=0 && Character.isDigit(text.charAt(i)))
			i--;
		if (i>0 && Character.isLetter(text.charAt(i))) return false;
		int number = Integer.parseInt(text.substring(i+1, j));
		return number >= RANGE_LOWER_BOUND && number <= RANGE_UPPER_BOUND;
	}
	
	/**
	 * Determines wheter the period at index <code>i</code> belongs to an acronym
	 * that is on the used acronym list.
	 * 
	 * @param text The text to be analysed
	 * @param i The index of the period
	 * @return <code>true</code> iff. period belongs to an acronym
	 */
	private boolean periodBelongsToAnAcronym(String text, int i){
		int begin = i-1;
		while(begin > 0 
				&& (Character.isLetter(text.charAt(begin)) 
						|| CharTools.isPeriod(text.charAt(begin))))
			begin--;
		return WordLists.isAcronym(text.substring(begin+1,i+1));	
	}
	
	
	
	// -------------------------------------------------------------------
	// DETERMINERS FOR SENTENCE BEGINS
	// -------------------------------------------------------------------
	
//	/**
//	 * Determines whether the next token is a probable start term of a sentence.
//	 * In the current implementation, it only checks if the next term begins with
//	 * an upper case letter.
//	 * 
//	 * @param text The text to be analysed
//	 * @param i The index of the current character
//	 * @return <code>true</code> iff. the next token is a probable sentence
//	 *   begin.
//	 */
//	private boolean successingTermIsSentenceBegin(String text, int i){
//		while (Character.isWhitespace(text.charAt(i+1)))
//				i++;
//		return characterMayBeSentenceBegin(text, i+1);
//	}
	
	/**
	 * Determines whether the next token is a probable start term of a sentence.
	 * In the current implementation, it only checks if the next term begins with
	 * an upper case letter.
	 * 
	 * @param text The text to be analysed
	 * @param i The index of the current character
	 * @return <code>true</code> iff. the next token is a probable sentence
	 *   begin.
	 */
	private boolean characterMayBeSentenceBegin(String text, int i){
		// Index is out of bounds
		if (i < 0 || i >= text.length())
			return false;
		// Check different possibilities
		char c = text.charAt(i);
		boolean quot = false;
		if (i < text.length()-6) quot = "&quot;".equals(text.substring(i, i+6));
		
		return Character.isUpperCase(c)
				|| quot
				|| CharTools.isQuotationMark(c)
				|| CharTools.isOpeningBracket(c)
				|| (CharTools.isTendencySymbol(c) && !CharTools.isTendencySymbol(text.charAt(i-1)));
	}

	/**
	 * Determines whether the next word after the character at position <CODE>i</CODE>
	 * is a commonly used word at the beginning of a sentence, i.e., it is a word
	 * starting with an upper-case letter, which is normally written in lower-case.
	 * This method is based on a word list of such words.
	 *  
	 * @param text The text to be analyzed
	 * @param i The index of the current character
	 * @return <CODE>true</CODE> iff. the next may be a sentence begin.
	 */
	private boolean nextWordMayBeSentenceBegin(String text, int i){
		if (i >= text.length()-1) return false;
		while (i < text.length()-1 && Character.isWhitespace(text.charAt(i)))
			i++;
		int begin = i;
		i++;
		while (i < text.length()-1 && Character.isLetter(text.charAt(i)))
			i++;
		return WordLists.isSentenceBegin(text.substring(begin, i));
	}
	
	/**
	 * Determines whether the token beginning at position <CODE>i</CODE>
	 * is a footnote of the form "[XY]".
	 *  
	 * @param text The text to be analyzed
	 * @param i The index of the current character
	 * @return <CODE>true</CODE> iff. the next token is a footnote of the form "[XY]"
	 */
	private boolean nextIsFootNote(String text, int i){
		if (i >= text.length()-1) return false;
		if (text.charAt(i) == '['){
			while ((i < text.length()-1 && text.charAt(i) != ']' && !Character.isWhitespace(text.charAt(i))))
				i++;
			if (text.charAt(i) == ']') return true;
		}
		return false;
	}
	
	
	
	// -------------------------------------------------------------------
	// DETERMINERS FOR LINE REQUESTS
	// -------------------------------------------------------------------
	
	/**
	 * Determines whether the next line after the character at position <CODE>i</CODE>
	 * starts with a possible bullet symbol as given in 
	 * {@link de.upb.infexba.util.CharTools}.
	 * 
	 * @param text The text to be analyzed
	 * @param i The index of the current character
	 * @return <CODE>true</CODE> iff. the next line starts with a bullet symbol
	 */
	private boolean nextLineStartsWithBullet(String text, int i){
		if (i >= text.length()-1) return false;
		while (i+1 < text.length() && Character.isWhitespace(text.charAt(i)))
			i++;
		return CharTools.isBullet(text.charAt(i));
	}
	
	/**
	 * Determines if the next non-whitespace character equals the character
	 * given by <code>c</code>.
	 * 
	 * @param text The text to be analysed
	 * @param c The character to be checked for equality
	 * @param i The index of the current character
	 * @return <code>true</code> iff. the two characters are equal
	 */
	private boolean nextLineStartsWithSameCharacter(String text, char c, int i){
		if (i >= text.length()-1) return false;
		int j = i;
		while (j+1<text.length() && Character.isWhitespace(text.charAt(j+1)))
			j++;
		if (j+1 == text.length()) return false;
		return c == text.charAt(j+1);
	}
	
	/**
	 * Determines whether the next line after the line break at index
	 * <code>i</code> is a blank line, i.e. it consists of an arbitrary
	 * amount of white spaces and nothing else.
	 * 
	 * @param text The text to be analysed
	 * @param i The index of the first line break
	 * @return <code>true</code> iff. the next line is a blank line
	 */
	private boolean nextLineIsBlankLine(String text, int i){
		// Document uses Carriage Return AND Line Feed
		if (text.charAt(i) == '\r' && text.charAt(i+1) == '\n'){
			int j = i+2;
			while (j< text.length() && Character.isSpaceChar(text.charAt(j)))
				j++;
			if (j >= text.length()) return true;
			return (text.charAt(j) == '\r' && text.charAt(j+1) == '\n');
		}
		// Document uses either Carriage Return OR Line Feed
		else if (text.charAt(i) == '\r' || text.charAt(i) == '\n'){
			char cur = text.charAt(i);
			int j = i+1;
			while (j< text.length() && Character.isSpaceChar(text.charAt(j)))
				j++;
			if (j >= text.length()) return true;
			return (text.charAt(j) == cur);
		}
		// Other cases not supported at the moment
		return false;
	}
	
//	/**
//	 * Determines whether the last line before the line break at index
//	 * <code>i</code> is a blank line, i.e. it consists of an arbitrary
//	 * amount of white spaces and nothing else.
//	 * 
//	 * @param text The text to be analysed
//	 * @param i The index of the first line break
//	 * @return <code>true</code> iff. the last line is a blank line
//	 */
//	private boolean previousLineIsBlankLine(String text, int i){
//		// Document uses Carriage Return AND Line Feed
//		if (i < 2) return true;
//		if (text.charAt(i-1) == '\n' && text.charAt(i-2) == '\r'){
//			int j = i-3;
//			while (j >= 0 && Character.isSpaceChar(text.charAt(j)))
//				j--;
//			if (j < 0) return true;
//			return (text.charAt(j) == '\n' && text.charAt(j-1) == '\r');
//		}
//		// Document uses either Carriage Return OR Line Feed
//		else if (text.charAt(i-1) == '\r' || text.charAt(i-1) == '\n'){
//			char cur = text.charAt(i-1);
//			int j = i-2;
//			while (j >= 0 && Character.isSpaceChar(text.charAt(j)))
//				j--;
//			if (j < 0) return true;
//			return (text.charAt(j) == cur);
//		}
//		// Other cases not supported at the moment
//		return false;
//	}
}
