package de.upb.infexba.ml;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.List;

/**
 * IO class for writing feature names and values to an ARFF
 * file.
 * 
 * @author hwachsmuth, 2010/10/26
 *
 */
public class GeneralFeatureARFFWriter {

	/**
	 * The title of the file to be written. The title is inserted
	 * into the header of the file.
	 */
	private String title;
	
	/**
	 * The name of the relation tag of the file to be written.
	 */
	private String relation;
	
	/**
	 * The possible classes written as a set string: "{class1, ..., classn}"
	 */
	private String classes;
	
	/**
	 * The constructor creates the object and sets the title and
	 * relation name.
	 * 
	 * @param title The title
	 * @param relation The relation
	 */
	public GeneralFeatureARFFWriter(String title, String relation, String classes){
		this.title = title;
		this.relation = relation;
		this.classes = classes;
	}
	
	/**
	 * The main method for writing the given set of feature instances with the given
	 * names to an ARFF file with the given location.
	 * 
	 * @param names The names
	 * @param instances The instance
	 * @param location The location
	 * @return <CODE>true</CODE> iff. file was successfully written
	 */
	public boolean writeFeaturesToARFF(List<String> names, List<RegularInstance> instances, 
			String location){
		try {
		    BufferedWriter out = new BufferedWriter(new FileWriter(location));
		    // Write header and data
		    this.writeHeader(names, out);
		    this.writeData(instances, out);
			// Some empty lines at the end
    		out.write("%\n");
    		out.write("%\n");
    		out.write("%\n");
    		out.close();
		} catch (IOException e) {
			return false;
		}
		return true;
	}

	/**
	 * Writes the header of an ARFF file including the given names to the given output
	 * writer.
	 * 
	 * @param names The names
	 * @param out The output writer
	 * @throws IOException Thrown if a problem occurs while writing to the writer
	 */
	private void writeHeader(List<String> names, BufferedWriter out) throws IOException{
	    // Write header
	    out.write("% 1. Title: "+ title + "\n");
		out.write("%\n");
		out.write("% 2. Sources:\n");
		out.write("%      (a) Creator: InfexBA\n");
		out.write("%      (c) Date: 2010\n");
		out.write("%\n\n");		
		out.write("@RELATION  " + relation + "\n\n");
		// Features
		if(names!=null)
			for (String name : names)
				out.write("@ATTRIBUTE " + name + "  NUMERIC\n");				
		out.write("@ATTRIBUTE class  " + classes + "\n\n");
	}
	
	/**
	 * Writes the instances of an ARFF file including the given names to the given output
	 * writer.
	 * 
	 * @param instances The instances
	 * @param out The output writer
	 * @throws IOException Thrown if a problem occurs while writing to the writer
	 */
	private void writeData(List<RegularInstance> instances, BufferedWriter out) throws IOException{
		out.write("@DATA\n");	
		if(instances != null)
			for (RegularInstance instance : instances) {
				for (Double value : instance.getFeatureValues())
					out.write(""+value + ",");
				out.write(instance.getClassification()+"\n");
			}
	}
}