package de.upb.infexba.ml;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.uima.cas.FSIterator;
import org.apache.uima.cas.text.AnnotationIndex;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;

import de.upb.infexba.Genre;
import de.upb.infexba.Sentence;
import de.upb.infexba.Token;
import de.upb.infexba.ml.feature.ngram.CharNgramHandler;
import de.upb.infexba.ml.feature.ngram.POSNgramHandler;
import de.upb.infexba.ml.feature.partofspeech.PartOfSpeechFeatureHandler;
import de.upb.infexba.ml.feature.vocabulary.CounterFeatureHandler;
import de.upb.infexba.ml.feature.vocabulary.VocabularyFeatureHandler;

/**
 * FeatureCalculator.
 * 
 * @author kabu
 * 
 */
public class IntentionFeatureCalculator {

	// -------------------------------------------------------------------
	// CONSTANTS
	// -------------------------------------------------------------------

	public static final String COMMERCIAL = "commercial";

	public static final String INFORMATIONAL = "informational";

	public static final String PERSONAL = "personal";

	// -------------------------------------------------------------------
	// Feature Handler
	// -------------------------------------------------------------------

	PartOfSpeechFeatureHandler posFeatureHandler;

	VocabularyFeatureHandler vocFeatureHandler;

	CounterFeatureHandler counterFeatureHandler;

	POSNgramHandler posNgramFeatureHandler;

	CharNgramHandler charNgramFeatureHandler;

	// -------------------------------------------------------------------
	// INITIALIZATION
	// -------------------------------------------------------------------

	private boolean get_first_annotation = true;

	/**
	 * IntentionFeatureCalculator.
	 * 
	 * @param pos
	 * @param counter
	 * @param word_key_skip_list
	 * @param explicit_word_count_list
	 */
	public IntentionFeatureCalculator(String[] pos, String[] counter, String[] word_key_skip_list, String[] lemma_key_skip_list,
			String[] pos_skip_list_for_tokens, boolean get_first_annotation) {

		this.vocFeatureHandler = new VocabularyFeatureHandler();
		this.vocFeatureHandler.setWordSkipList(word_key_skip_list);
		this.vocFeatureHandler.setLemmaSkipList(lemma_key_skip_list);
		this.vocFeatureHandler.setPOSSkipList(pos_skip_list_for_tokens);

		this.posFeatureHandler = new PartOfSpeechFeatureHandler(pos);

		this.counterFeatureHandler = new CounterFeatureHandler(counter);

		this.posNgramFeatureHandler = new POSNgramHandler(3);
		this.charNgramFeatureHandler = new CharNgramHandler(3);

		this.get_first_annotation = get_first_annotation;
	}

	public VocabularyFeatureHandler getVocabularyFeatureHandler() {
		return this.vocFeatureHandler;
	}

	public PartOfSpeechFeatureHandler getPartOfSpeechFeatureHandler() {
		return this.posFeatureHandler;
	}

	public CounterFeatureHandler getCounterFeatureHandler() {
		return this.counterFeatureHandler;
	}

	public POSNgramHandler getPOSNgramFeatureHandler() {
		return this.posNgramFeatureHandler;
	}

	public CharNgramHandler getCharNgramFeatureHandler() {
		return this.charNgramFeatureHandler;
	}

	private void clearFeatureMapsOfAllHandlers() {
		vocFeatureHandler.clearFeatureMap();
		posFeatureHandler.clearFeatureMap();
		counterFeatureHandler.clearFeatureMap();
		posNgramFeatureHandler.clearFeatureMap();
		charNgramFeatureHandler.clearFeatureMap();
	}

	private void clearFeaturesOfAllHandlers() {
		this.posFeatureHandler.clearFeatures();
		this.vocFeatureHandler.clearFeatures();
		this.counterFeatureHandler.clearFeatures();
		this.posNgramFeatureHandler.clearFeatures();
		this.charNgramFeatureHandler.clearFeatures();
	}
	
	public void setNormalizationMode(short mode){
		this.posFeatureHandler.setNormalizationMode(mode);
		this.vocFeatureHandler.setNormalizationMode(mode);
		this.counterFeatureHandler.setNormalizationMode(mode);
		this.posNgramFeatureHandler.setNormalizationMode(mode);
		this.charNgramFeatureHandler.setNormalizationMode(mode);
	}

	// -------------------------------------------------------------------
	// HELPER-METHODS for protocols
	// -------------------------------------------------------------------

	/**
	 * @return
	 */
	public String getInfo() {
		String s = "\n-----------------------------------\n" + "Paramters of the Feature Calculator:" + "\n-----------------------------------\n\n";
		s += string_set_map_to_string(vocFeatureHandler.string_set_features);
		s += "\n\nSkip-Lists:\n\n";
		s += "lemma key skip list = " + set_to_string(this.vocFeatureHandler.getWordKeySkipList()) + "\n\n";
		s += "word key skip list = " + set_to_string(this.vocFeatureHandler.getLemmaKeySkipList());
		return s;
	}

	// -------------------------------------------------------------------
	// FEATURE CALCULATION
	// -------------------------------------------------------------------

	/**
	 * Computes the features values for the features listed in featureNames or just saves information about all features.
	 * 
	 * If no featureNames are defined (i.e. featureNames is null), then all known features will be computed and information about them will be stored in the
	 * specific maps. If featureNames are given, then those featureNames will be computed and a vector with the values will be returned.
	 * 
	 * @param jcas
	 * @param featureNames
	 * @return
	 */
	public List<Double> computeFeatureValues(JCas jcas, List<String> featureNames) {

		boolean isTraining = (featureNames == null);
		String classification = null;

		if (!isTraining) {

			// feature names are given, so tell the handlers to handle exactly these
			this.clearFeaturesOfAllHandlers();
			this.clearFeatureMapsOfAllHandlers();
			for (String feature : featureNames)
				if (posFeatureHandler.canHandle(feature))
					posFeatureHandler.addFeature(feature);
				else if (counterFeatureHandler.canHandle(feature))
					counterFeatureHandler.addFeature(feature);
				else if (vocFeatureHandler.canHandle(feature))
					vocFeatureHandler.addFeature(feature);
				else if (posNgramFeatureHandler.canHandle(feature))
					posNgramFeatureHandler.addFeature(feature);
				else if (charNgramFeatureHandler.canHandle(feature))
					charNgramFeatureHandler.addFeature(feature);

		} else {

			// if no features are given, it's not necessary to tell the handlers anything,
			// but determine classification
			classification = getClassificationByAnnotation(jcas, this.get_first_annotation);
		}

		// Iterators
		AnnotationIndex<Annotation> sentenceIndex = jcas.getAnnotationIndex(Sentence.type);
		AnnotationIndex<Annotation> tokenIndex = jcas.getAnnotationIndex(Token.type);
		FSIterator<Annotation> tokenIterator;
		FSIterator<Annotation> sentenceIterator = sentenceIndex.iterator();
		Sentence sentence;
		Token token;

		while (sentenceIterator.hasNext()) {
			sentence = (Sentence) sentenceIterator.next();
			
			// counts per sentence
			counterFeatureHandler.handle(sentence);

			tokenIterator = tokenIndex.subiterator(sentence);
			while (tokenIterator.hasNext()) {

				token = (Token) tokenIterator.next();
				String word = token.getCoveredText();
				String lemma = token.getLemma();

				// part-of-speech
				String pos = PartOfSpeechFeatureHandler.getPos(token);
				posFeatureHandler.handle(pos);

				// words/lemma
				vocFeatureHandler.handle(word, lemma, pos, classification);

				// counter per word
				counterFeatureHandler.handle(word);
			}
		}

		// IDF-maps: stores in how many documents a word appears
		if (isTraining) {
			vocFeatureHandler.training_save_occurences();
			vocFeatureHandler.training_count_classification(classification);
		}

		// counts e.g. emoticons, moneyExpressions,...
		counterFeatureHandler.handle(jcas);
		// computes features from counts
		counterFeatureHandler.computeFeatures();

		// N-grams
		posNgramFeatureHandler.handle(jcas, classification);
		charNgramFeatureHandler.handle(jcas, classification);

		// Create feature-vector according to names given in feature-names
		// or Save the information
		List<Double> featureValues = new ArrayList<Double>();
		String feature;
		if (!isTraining) {
			// store feature-vector
			for (int i = 0; i < featureNames.size(); i++) {
				feature = featureNames.get(i);
				if (posFeatureHandler.handles(feature))
					featureValues.add(posFeatureHandler.getFromFeatureMap(feature));
				else if (counterFeatureHandler.handles(feature))
					featureValues.add(counterFeatureHandler.getFromFeatureMap(feature));
				else if (vocFeatureHandler.handles(feature))
					featureValues.add(vocFeatureHandler.getFromFeatureMap(feature));
				else if (posNgramFeatureHandler.handles(feature))
					featureValues.add((double) posNgramFeatureHandler.getFromFeatureMap(feature));
				else if (charNgramFeatureHandler.handles(feature))
					featureValues.add((double) charNgramFeatureHandler.getFromFeatureMap(feature));
				else{
					System.err.println("Feature "+feature+" not handled by any featureHandler.");
					featureValues.add(0.);
				}
			}
			// clear
			clearFeatureMapsOfAllHandlers();

		} // else no feature names given, so store everything and do not clear anything

		return featureValues;
	}

	public double getNormalizedValue(String feature, double value, int numberOfTokens) {
		if (this.getCounterFeatureHandler().canHandle(feature))
			value = this.getCounterFeatureHandler().getNormalizedValue(feature, value, null);
		else if (this.getPartOfSpeechFeatureHandler().canHandle(feature))
			value = this.getPartOfSpeechFeatureHandler().getNormalizedValue(feature, value, (double) numberOfTokens);
		else if (this.getVocabularyFeatureHandler().canHandle(feature))
			value = this.getVocabularyFeatureHandler().getNormalizedValue(feature, value, (double) numberOfTokens);
		else if (this.getCharNgramFeatureHandler().canHandle(feature))
			value = this.getCharNgramFeatureHandler().getNormalizedValue(feature, value, (double) numberOfTokens);
		else if (this.getPOSNgramFeatureHandler().canHandle(feature))
			value = this.getPOSNgramFeatureHandler().getNormalizedValue(feature, value, (double) numberOfTokens);
		return value;
	}

	// -------------------------------------------------------------------
	// FEATURE-NAME/TYPE-HANDLING
	// -------------------------------------------------------------------

	private static String set_to_string(Set<String> list) {
		String s = "";
		int counter = 0;
		for (String l : list) {
			s += l + ", ";
			counter++;
			if (counter % 20 == 0)
				s += "\n";
		}
		return s;
	}

	private static String string_set_map_to_string(HashMap<String, HashSet<String>> set) {
		String s = "";
		for (String key : set.keySet()) {
			s = s + key + " = ";
			HashSet<String> string_set = set.get(key);
			for (String f : string_set)
				s = s + f + ", ";
			s = s + "\n";
		}
		return s;
	}

	/**
	 * Returns the FIRST Classification of the JCas Object saved in its annotation.
	 * 
	 * @param jcas
	 * @return
	 */
	public static String getClassificationByAnnotation(JCas jcas, boolean get_first_annotation) {
		String classification = "NONE";
		FSIterator<Annotation> genreIter = jcas.getAnnotationIndex(Genre.type).iterator();

		if (get_first_annotation) {
			if (genreIter.hasNext()) {
				Annotation a = genreIter.next();
				if (a.getTypeIndexID() == Genre.type) {
					classification = ((Genre) a).getGenreClass();
					// System.err.println("classification = "+classification);
				}
			}
		} else {
			while (genreIter.hasNext()) {
				Annotation a = genreIter.next();
				if (a.getTypeIndexID() == Genre.type) {
					classification = ((Genre) a).getGenreClass();
					// System.err.println("classification = "+classification);
				}
			}
		}
		return classification;
	}
}
