package de.upb.infexba.ml;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;
import java.util.TreeSet;

import org.apache.uima.cas.FSIterator;
import org.apache.uima.cas.text.AnnotationIndex;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;

import de.upb.infexba.Token;
import de.upb.infexba.ml.feature.ngram.CharNgramHandler;
import de.upb.infexba.ml.feature.ngram.POSNgramHandler;
import de.upb.infexba.ml.feature.partofspeech.PartOfSpeechFeatureHandler;
import de.upb.infexba.ml.feature.vocabulary.CounterFeatureHandler;
import de.upb.infexba.ml.feature.vocabulary.SkipLists;
import de.upb.infexba.ml.feature.vocabulary.VocabularyFeatureHandler;

/**
 * Dummy class for generating features to train and user a document intention classifier.
 * 
 * @author kabu
 * 
 */
public class IntentionFeatureGenerator {

	// -------------------------------------------------------------------
	// CONSTANTS
	// -------------------------------------------------------------------

	/**
	 * The name of the intention class commercial
	 */
	public static final String COMMERCIAL = "commercial";

	/**
	 * The name of the intention class informational
	 */
	public static final String INFORMATIONAL = "informational";

	/**
	 * The name of the intention class personal
	 */
	public static final String PERSONAL = "personal";

	// -------------------------------------------------------------------
	// VARIABLES
	// -------------------------------------------------------------------

	/**
	 * Results of the computeFeatures method (for all JCAS objects)
	 */
	private List<String> computedFeatureNames;

	private List<RegularInstance> computedFeatureValues;

	// ---------------------------------------------------------------------------------------------
	//
	// ---------------------------------------------------------------------------------------------

	private IntentionFeatureCalculator featureCalc;

	static boolean get_first_annotation = true;

	public void setGetFirstAnnotation(boolean getFirst) {
		get_first_annotation = getFirst;
	}

	static long seed = 321353169;

	public void setSeed(long seed) {
		IntentionFeatureGenerator.seed = seed;
	}

	// ---------------------------------------------------------------------------------------------
	// OPTIONS
	// ---------------------------------------------------------------------------------------------

	// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	// POS and SKIP-LISTS
	// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	String[] POS = PartOfSpeechFeatureHandler.ALL_AND_GROUPS; // POS-Tags and -Groups

	String[] COUNTER = CounterFeatureHandler.COUNTER_ALL; // { CounterFeatureHandler.TIME_EXPR }; //

	String[] WORD_KEY_SKIP_LIST = SkipLists.WORD_KEY_SKIP_LIST_SPECIFIC;

	String[] LEMMA_KEY_SKIP_LIST = SkipLists.LEMMA_KEY_SKIP_LIST_MUSIC;

	String[] POS_SKIP_LIST_FOR_TOKEN = SkipLists.POS_SKIP_LIST_FOR_TOKENS;

	String[] EXPLICIT_WORD_COUNT_LIST = {}; // FeatureCalculator.SATZZEICHEN;

	// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	// FEATURES TO BE COMPUTED
	// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	boolean ADD_GS_AS_SETS = false; // if this is selected, the other gs options are disabled!
	boolean IF_ADD_GS_AS_SETS_THEN_ADDITIONALLY_SELF_DEFED = false;

	// Explicit
	boolean ONLY_USE_EXPLICIT_WORD_LIST = false;

	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// ... Word/Lemma
	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	// Set to true, if you want you want to use the inverse document frequency to compute genre-specific
	// word/lemma-features. Otherwise a simple ratio will be
	// computed.
	boolean NORMALIZE_GS_RATIO__IDF = false;

	// Set to true, if you want to normalize the simple ratio by the text-length, i.e. the ratio of occurrences of the
	// words will be multiplied by the (inverse)
	// ratio of the text-lengths.
	boolean NORMALIZE_GS_RATIO__TEXTLENGTH = false; // normalization w.r.t number of tokens

	// Top-Lemma?
	boolean GET_TOP_LEMMA = false;
	// ... parameters:
	int GET_TOP_LEMMA_T = 10;

	// Genre-Specific-Lemma?
	boolean GET_GENRE_SPECIFIC_LEMMA = false;
	// ... parameters:
	int GET_GENRE_SPECIFIC_LEMMA_MIN = 15;
	double GET_GENRE_SPECIFIC_LEMMA_THRESHOLD = 2;
	int GET_GENRE_SPECIFIC_LEMMA_T = 30;

	// Top-Words?
	boolean GET_TOP_WORDS = true;
	// ... parameters:
	int GET_TOP_WORDS_T = 40;
	int SKIP_THE_FIRST_X_TOP_WORDS = 0; // !

	// Genre-Specific-Words?
	boolean GET_GENRE_SPECIFIC_WORDS = true;
	// .... parameters:
	int GET_GENRE_SPECIFIC_WORDS_T = 40;
	int GET_GENRE_SPECIFIC_WORDS_MIN = 15; // problem: commercial corpus is small... => 10
	double GET_GENRE_SPECIFIC_WORDS_THRESHOLD = 6; // ... and informational corpus hasn't got many words that appear
	// more often => 8

	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// ... NGram
	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	// Top Char-N-Grams?
	boolean GET_TOP_CHARNGRAMS = true;
	int GET_TOP_CHARNGRAMS_T = 80; // Integer.MAX_VALUE; //

	// Genre Specific Char-N-Grams?
	boolean GET_GENRE_SPECIFIC_CHARNGRAMS = true;
	int GET_GENRE_SPECIFIC_CHARNGRAMS_T = 40;
	int GET_GENRE_SPECIFIC_CHARNGRAMS_MIN = 40;
	int GET_GENRE_SPECIFIC_CHARNGRAMS_THRESHOLD = 6;

	// Top POS-N-Grams?
	boolean GET_TOP_POSNGRAMS = true;
	int GET_TOP_POSNGRAMS_T = 80; // Integer.MAX_VALUE; //

	// Genre Specific POS-N-Grams?
	boolean GET_GENRE_SPECIFIC_POSNGRAMS = true;
	int GET_GENRE_SPECIFIC_POSNGRAMS_T = 40;
	int GET_GENRE_SPECIFIC_POSNGRAMS_MIN = 15;
	int GET_GENRE_SPECIFIC_POSNGRAMS_THRESHOLD = 6;

	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// ... POS
	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	int LOWER_BOUND_FOR_POS_OCCURRENCES = 15;

	// ---------------------------------------------------------------------------------------------
	// METHODS
	// ---------------------------------------------------------------------------------------------

	public IntentionFeatureGenerator() {
		featureCalc = new IntentionFeatureCalculator(POS, COUNTER, WORD_KEY_SKIP_LIST, LEMMA_KEY_SKIP_LIST, POS_SKIP_LIST_FOR_TOKEN, get_first_annotation);
		computedFeatureValues = new ArrayList<RegularInstance>();
		computedFeatureNames = new ArrayList<String>();
	}

	/**
	 * This method should be used to find features based on the given input document, such as which part-of-speech appears and so on. It is only called for the
	 * training set.
	 * 
	 * 
	 * @param jcas The JCas object - it gives you access to all the syntactic and semantic information known about the document.
	 */
	public void initializeFeatureSet(JCas jcas, int number_of_trainingsfiles) {

		// Initially all features the FeatureCalculator knows about will be calculated and information
		// about this features will be stored internally in the featureCalculator to reduce the set of
		// features later (see: reduceFeatureSet).
		featureCalc.computeFeatureValues(jcas, null);
	}

	/**
	 * In this method, features that were found in <CODE>initializeFeatureSet</CODE>, but are quite rare and, thus, might fail statistical significance, can be
	 * deleted.
	 * 
	 * Returns information about features and setting.
	 * 
	 */
	public String reduceFeatureSet() {

		// ... in fact we add features to computedFeatureNames and just leave out rare features.

		TreeSet<String> set = new TreeSet<String>();

		// *************************************************************************************************
		// COUNTER-features

		println("\n---------------------------------------------------------------------------------------------------");
		println("COUNTER:\n");

		println("added COUNTER-features: ");
		for (String feature : featureCalc.getCounterFeatureHandler().getFeatures()) {
			if (featureCalc.getCounterFeatureHandler().getFromFeatureMap(feature) > 0)
				set.add(feature);
			else
				System.err.println("o.O das feature " + feature + " wurde kein mal gezählt!");
		}
		addFeaturesToComputedFeatures(set, null, true);
		set.clear();

		// *************************************************************************************************
		// POS-features

		println("\n---------------------------------------------------------------------------------------------------");
		println("POS:\n");

		println("added POS-features: ");
		for (String feature : featureCalc.getPartOfSpeechFeatureHandler().getFeatures()) {
			if (featureCalc.getPartOfSpeechFeatureHandler().getFromFeatureMap(feature) > LOWER_BOUND_FOR_POS_OCCURRENCES)
				set.add(feature);
		}
		addFeaturesToComputedFeatures(set, null, true);
		set.clear();

		// *************************************************************************************************
		// NGRAM-features

		Set<String> all_feature_strings = new HashSet<String>(); // to avoid duplicate features

		println("\n---------------------------------------------------------------------------------------------------");
		println("NGRAMs:");

		if (GET_GENRE_SPECIFIC_CHARNGRAMS) {
			println("\nadded GenreSpecific CHAR-NGRAM-features: ");
			addFeaturesToComputedFeatures(
					this.featureCalc.getCharNgramFeatureHandler().getAllGenreSpecific(GET_GENRE_SPECIFIC_CHARNGRAMS_T, GET_GENRE_SPECIFIC_CHARNGRAMS_THRESHOLD,
							GET_GENRE_SPECIFIC_CHARNGRAMS_MIN, false), all_feature_strings, false);
		}
		if (GET_TOP_CHARNGRAMS) {
			println("\nadded Top CHAR-NGRAM-features: ");
			addFeaturesToComputedFeatures(this.featureCalc.getCharNgramFeatureHandler().getTop(GET_TOP_CHARNGRAMS_T, 0), all_feature_strings, false);
		}

		print("\n");

		if (GET_GENRE_SPECIFIC_POSNGRAMS) {
			println("added GenreSpecific POS-NGRAM-features: ");
			addFeaturesToComputedFeatures(
					this.featureCalc.getPOSNgramFeatureHandler().getAllGenreSpecific(GET_GENRE_SPECIFIC_POSNGRAMS_T, GET_GENRE_SPECIFIC_POSNGRAMS_THRESHOLD,
							GET_GENRE_SPECIFIC_POSNGRAMS_MIN, false), all_feature_strings, false);
		}
		if (GET_TOP_POSNGRAMS) {
			println("added Top POS-NGRAM-features: ");
			addFeaturesToComputedFeatures(this.featureCalc.getPOSNgramFeatureHandler().getTop(GET_TOP_POSNGRAMS_T, 0), all_feature_strings, false);
		}

		print("\n");

		// *************************************************************************************************
		// STRING-features

		println("\n---------------------------------------------------------------------------------------------------");
		println("WORD/LEMMA:\n");

		if (ONLY_USE_EXPLICIT_WORD_LIST && EXPLICIT_WORD_COUNT_LIST.length > 0) {

			println("\nUse Only SelfDefined Words: ");

			set = new TreeSet<String>(Arrays.asList(EXPLICIT_WORD_COUNT_LIST));
			addFeaturesToComputedFeatures(set, all_feature_strings, true);
			set.clear();

		} else {

			if (ADD_GS_AS_SETS) {

				println("\nUse word-sets as features");

				int min = GET_GENRE_SPECIFIC_WORDS_MIN;
				double threshold = GET_GENRE_SPECIFIC_WORDS_THRESHOLD;
				int T = GET_GENRE_SPECIFIC_WORDS_T;

				HashSet<String> personal_specific_words = new HashSet<String>();
				HashSet<String> informational_specific_words = new HashSet<String>();
				HashSet<String> commercial_specific_words = new HashSet<String>();
				if (NORMALIZE_GS_RATIO__IDF) {
					personal_specific_words.addAll(featureCalc.getVocabularyFeatureHandler().get_genre_specific_words_for_one_class__idf(T, threshold,
							PERSONAL, COMMERCIAL, INFORMATIONAL));
					informational_specific_words.addAll(featureCalc.getVocabularyFeatureHandler().get_genre_specific_words_for_one_class__idf(T, threshold,
							INFORMATIONAL, PERSONAL, COMMERCIAL));
					commercial_specific_words.addAll(featureCalc.getVocabularyFeatureHandler().get_genre_specific_words_for_one_class__idf(T, threshold,
							COMMERCIAL, PERSONAL, INFORMATIONAL));
				} else {
					personal_specific_words.addAll(featureCalc.getVocabularyFeatureHandler().getGenreSpecificForOneClass(T,
							VocabularyFeatureHandler.WORD_FEATURE, PERSONAL, threshold, min, NORMALIZE_GS_RATIO__TEXTLENGTH));
					informational_specific_words.addAll(featureCalc.getVocabularyFeatureHandler().getGenreSpecificForOneClass(T,
							VocabularyFeatureHandler.WORD_FEATURE, INFORMATIONAL, threshold, min, NORMALIZE_GS_RATIO__TEXTLENGTH));
					commercial_specific_words.addAll(featureCalc.getVocabularyFeatureHandler().getGenreSpecificForOneClass(T,
							VocabularyFeatureHandler.WORD_FEATURE, COMMERCIAL, threshold, min, NORMALIZE_GS_RATIO__TEXTLENGTH));
				}

				// avoid duplicates
				HashSet<String> personal_specific_words2 = new HashSet<String>();
				HashSet<String> informational_specific_words2 = new HashSet<String>();
				HashSet<String> commercial_specific_words2 = new HashSet<String>();
				for (String s : personal_specific_words)
					if (!informational_specific_words.contains(s) && !commercial_specific_words.contains(s))
						personal_specific_words2.add(s);
				for (String s : informational_specific_words)
					if (!personal_specific_words.contains(s) && !commercial_specific_words.contains(s))
						informational_specific_words2.add(s);
				for (String s : commercial_specific_words)
					if (!informational_specific_words.contains(s) && !personal_specific_words.contains(s))
						commercial_specific_words2.add(s);

				computedFeatureNames.add(featureCalc.getVocabularyFeatureHandler().define_as_one_feature(personal_specific_words));
				computedFeatureNames.add(featureCalc.getVocabularyFeatureHandler().define_as_one_feature(informational_specific_words));
				computedFeatureNames.add(featureCalc.getVocabularyFeatureHandler().define_as_one_feature(commercial_specific_words));

				if (IF_ADD_GS_AS_SETS_THEN_ADDITIONALLY_SELF_DEFED) {
					println("Additionally add SelfDefined Words: ");
					set = new TreeSet<String>(Arrays.asList(EXPLICIT_WORD_COUNT_LIST));
					addFeaturesToComputedFeatures(set, all_feature_strings, true);
					set.clear();
				}

			} else { // add single string features

				all_feature_strings = new HashSet<String>(); // to avoid duplicate features

				// (a) lemma

				if (GET_GENRE_SPECIFIC_LEMMA) {
					println("\nadded genre-specific-LEMMA-features: ");
					if (NORMALIZE_GS_RATIO__IDF)
						addFeaturesToComputedFeatures(
								featureCalc.getVocabularyFeatureHandler().get_all_genre_specific_words__idf(GET_GENRE_SPECIFIC_LEMMA_T,
										GET_GENRE_SPECIFIC_LEMMA_THRESHOLD), all_feature_strings, false);
					else
						addFeaturesToComputedFeatures(
								featureCalc.getVocabularyFeatureHandler().getAllGenreSpecific(GET_GENRE_SPECIFIC_LEMMA_T,
										VocabularyFeatureHandler.LEMMA_FEATURE, GET_GENRE_SPECIFIC_LEMMA_THRESHOLD, GET_GENRE_SPECIFIC_LEMMA_MIN,
										NORMALIZE_GS_RATIO__TEXTLENGTH), all_feature_strings, false);
				}

				if (GET_TOP_LEMMA) {
					addFeaturesToComputedFeatures(featureCalc.getVocabularyFeatureHandler().getTop(GET_TOP_LEMMA_T, VocabularyFeatureHandler.LEMMA_FEATURE, 0),
							all_feature_strings, false);
					println("\nadded top-LEMMA-features: ");
				}

				// (b) word

				if (GET_GENRE_SPECIFIC_WORDS) {
					println("\nadded genre-specific-WORD-features: ");
					if (NORMALIZE_GS_RATIO__IDF)
						addFeaturesToComputedFeatures(
								featureCalc.getVocabularyFeatureHandler().get_all_genre_specific_words__idf(GET_GENRE_SPECIFIC_WORDS_T,
										GET_GENRE_SPECIFIC_WORDS_THRESHOLD), all_feature_strings, false);
					else
						addFeaturesToComputedFeatures(
								featureCalc.getVocabularyFeatureHandler().getAllGenreSpecific(GET_GENRE_SPECIFIC_WORDS_T,
										VocabularyFeatureHandler.WORD_FEATURE, GET_GENRE_SPECIFIC_WORDS_THRESHOLD, GET_GENRE_SPECIFIC_WORDS_MIN,
										NORMALIZE_GS_RATIO__TEXTLENGTH), all_feature_strings, false);
				}

				if (GET_TOP_WORDS) {
					println("\nadded top-WORD-features: ");
					addFeaturesToComputedFeatures(
							featureCalc.getVocabularyFeatureHandler()
									.getTop(GET_TOP_WORDS_T, VocabularyFeatureHandler.WORD_FEATURE, SKIP_THE_FIRST_X_TOP_WORDS), all_feature_strings, false);
				}
			}
		}

		return getSetting() + info + featureCalc.getInfo();
	}

	/**
	 * Add all features from words-set to computedFeatureNames that are _not_ already contained in the all_feature_words (and then add it to the
	 * all_feature_words).
	 * 
	 * @param words a set of word-keys!
	 * @param all_feature_words
	 */
	private void addFeaturesToComputedFeatures(Set<String> features, Set<String> all_feature_words, boolean printThem) {
		if (features.isEmpty())
			return;

		int counter = 0;
		for (String feature : features) {

			String key = feature;
			if (this.featureCalc.getVocabularyFeatureHandler().canHandle(feature))
				key = VocabularyFeatureHandler.getTextFromFeatureName(feature);
			else if (this.featureCalc.getCharNgramFeatureHandler().canHandle(feature))
				key = CharNgramHandler.getTextFromFeatureName(feature);
			else if (this.featureCalc.getPOSNgramFeatureHandler().canHandle(feature))
				key = POSNgramHandler.getTextFromFeatureName(feature);

			if (all_feature_words != null) {
				if (!all_feature_words.contains(key)) {
					computedFeatureNames.add(feature);
					all_feature_words.add(key);
				}
			} else
				computedFeatureNames.add(feature);

			if (printThem) {
				print("\"" + key + "\", ");
				counter++;
				if (counter % 20 == 0)
					print("\n");
			}
		}
		if (printThem)
			print("\n");
	}

	/**
	 * This method is called for datasets different from the training set. It should be used to create features for all the given names since it is called after
	 * reading a feature set from a file.
	 * 
	 * @param featureNames The names of the features to be created.
	 */
	public void setFeatureSet(List<String> featureNames) {
		computedFeatureNames = featureNames;
	}

	/**
	 * In this method, a feature vector together with the intention class should be computed for the document given in the JCas object. This vector should be
	 * stored somewhere internally.
	 * 
	 * Don't forget to normalize you feature values!
	 * 
	 * Notice that the intention class of the current document can be accessed via the JCas object.
	 * 
	 * @param jcas The JCas object
	 */
	public void computeFeatures(JCas jcas) {
		// Compute the feature values.
		List<Double> featureValues = featureCalc.computeFeatureValues(jcas, computedFeatureNames);

		// Normalize the feature values.
		normalizeFeatureSet(jcas, featureValues);

		// Get the classification from the jCas object.
		String classification = IntentionFeatureCalculator.getClassificationByAnnotation(jcas, get_first_annotation);

		// Store the feature vector and the classification.
		if (featureValues == null)
			System.out.println("feature Values == null");
		if (computedFeatureValues == null)
			System.out.println("computedFeatureValues == null");

		computedFeatureValues.add(new RegularInstance(featureValues, classification));
	}

	public void setNormalizationMode(short mode) {
		this.featureCalc.setNormalizationMode(mode);
	}

	public void normalizeFeatureSet(JCas jcas, List<Double> featureValues) {
		String feature;
		Double value;
		// get the number of tokens
		int number_of_tokens = 0;
		AnnotationIndex<Annotation> tokenIndex = jcas.getAnnotationIndex(Token.type);
		FSIterator<Annotation> tokenIterator = tokenIndex.iterator();
		while (tokenIterator.hasNext()) {
			tokenIterator.next();
			number_of_tokens++;
		}
		// normalize the values
		for (int index = 0; index < computedFeatureNames.size(); index++) {
			feature = computedFeatureNames.get(index);
			value = featureValues.get(index);
			// stretch the values by individual factors
			value = featureCalc.getNormalizedValue(feature, value, number_of_tokens);
			// catch values exceeding [0,1]
			if (value < 0)
				value = 0.;
			if (value > 1)
				value = 1.;
			// store normalized value
			featureValues.set(index, value);
		}
	}

	/**
	 * This method differs from the other <CODE>computeFeatures</CODE> method in that it is used at classification time. That is, no classification is known,
	 * only the feature vector shall be computed and returned.
	 * 
	 * Don't forget to normalize you feature values!
	 * 
	 * The given annotation defines on which part of the input document the feature values shall be computed.
	 * 
	 * @param jcas The JCas object
	 * @param annotation The annotation
	 * @return The feature vector
	 */
	public double[] computeFeatures(JCas jcas, Annotation annotation) {
		// TODO: annotation...
		List<Double> features = featureCalc.computeFeatureValues(jcas, computedFeatureNames);
		double[] values = new double[features.size()];
		for (int i = 0; i < values.length; i++) {
			values[i] = features.get(i);
		}
		// annotation.getCoveredText();
		return values;
	}

	/**
	 * This method shall balance the internally stored set of training instances. If the parameter is set to <CODE>true</CODE>, random undersampling shall be
	 * used, else random duplicate oversampling shall be used.
	 * 
	 * Balancing means that there is exactly the same number of instances for each intention class afterwards.
	 * 
	 * @param doUndersampling
	 */
	@SuppressWarnings("unchecked")
	public void balance(boolean doUndersampling) {

		// First sort the classes and store instances in one list per classification.
		int[] counter = new int[3];
		List<RegularInstance>[] instances_per_class = new ArrayList[3];
		for (int c = 0; c < 3; c++) {
			instances_per_class[c] = new ArrayList<RegularInstance>();
		}
		RegularInstance ri;
		for (int i = 0; i < computedFeatureValues.size(); i++) {
			ri = computedFeatureValues.get(i);
			int c = 0;
			if (ri.getClassification().equals(COMMERCIAL))
				c = 0;
			else if (ri.getClassification().equals(INFORMATIONAL))
				c = 1;
			else
				c = 2;
			counter[c]++;

			instances_per_class[c].add(ri);
		}

		System.out.println("commercial = " + instances_per_class[0].size());
		System.out.println("informational = " + instances_per_class[1].size());
		System.out.println("personal = " + instances_per_class[2].size());

		// Determine how many instances should be in each class.
		int number_per_class = Math.max(counter[0], Math.max(counter[1], counter[2]));
		if (doUndersampling)
			number_per_class = Math.min(counter[0], Math.min(counter[1], counter[2]));

		// Randomly add instances to computedFeatureValues or randomly delete instances from the
		// instances_per_class lists.

		Random random = new Random(seed); // fixed seed

		int diff = 0;
		int index = 0;
		for (int c = 0; c < 3; c++) {
			if (doUndersampling)
				diff = counter[c] - number_per_class;
			else
				diff = number_per_class - counter[c];

			for (int i = 0; i < diff; i++) {
				index = (int) Math.floor(random.nextDouble() * instances_per_class[c].size());

				if (instances_per_class[c].isEmpty())
					System.err.println("class " + c + " is empty :( (0=commercial, 1=informational, 2=personal)");
				else {

					if (doUndersampling)
						instances_per_class[c].remove(index);
					else
						computedFeatureValues.add(instances_per_class[c].get(index));
				}
			}
		}

		// If we deleted instances from the instances_per_class lists, put them back together to the
		// computedFeatureValues list.
		if (doUndersampling) {
			computedFeatureValues = new ArrayList<RegularInstance>();
			for (int c = 0; c < 3; c++)
				computedFeatureValues.addAll(instances_per_class[c]);
		}
	}

	/**
	 * Returns the list of feature names.
	 * 
	 * @return The list of feature names
	 */
	public List<String> getFeatureNames() {
		return computedFeatureNames;
	}

	/**
	 * Returns the computed features
	 * 
	 * @return The computed features
	 */
	public List<RegularInstance> getFeatures() {
		return computedFeatureValues;
	}

	public void clear() {
		computedFeatureValues.removeAll(computedFeatureValues);
	}

	private static void print(String s) {
		System.out.print(s);
	}

	private static void println(String s) {
		System.out.println(s);
	}

	// ---------------------------------------------------------------------------------------------
	// FOR PROTOCOL
	// ---------------------------------------------------------------------------------------------

	private String info = "";

	String getSetting() {
		String setting = "----------------------------------------------\n" + "Paramters of the Intention Feature Generator:\n"
				+ "----------------------------------------------\n\n";
		if (GET_TOP_LEMMA)
			setting += "top lemmas, ";
		if (GET_GENRE_SPECIFIC_LEMMA)
			setting += "genre specific lemma, ";
		if (GET_TOP_WORDS)
			setting += "top words, ";
		if (GET_GENRE_SPECIFIC_WORDS)
			setting += "genre specific words, ";
		if (ADD_GS_AS_SETS)
			setting += "genre specific words as sets ";
		if (IF_ADD_GS_AS_SETS_THEN_ADDITIONALLY_SELF_DEFED)
			setting += "and additionally self defined words, ";
		else if (ADD_GS_AS_SETS)
			setting += ", ";
		setting += "\n";
		if (GET_GENRE_SPECIFIC_CHARNGRAMS)
			setting += "gs CHAR-ngrams, ";
		if (GET_GENRE_SPECIFIC_POSNGRAMS)
			setting += "gs POS-ngrams, ";
		if (GET_TOP_POSNGRAMS)
			setting += "top POS-ngrams, ";
		if (GET_TOP_CHARNGRAMS)
			setting += "top CHAR-ngrams, ";
		setting += "\n\n";

		setting += "ADD_GS_AS_SETS = " + ADD_GS_AS_SETS + "\n";
		setting += "IF_ADD_GS_AS_SETS_THEN_ADDITIONALLY_SELF_DEFED = " + IF_ADD_GS_AS_SETS_THEN_ADDITIONALLY_SELF_DEFED + "\n";
		setting += "ONLY_USE_EXPLICIT_WORD_LIST = " + ONLY_USE_EXPLICIT_WORD_LIST + "\n";

		setting += "\n\n";

		setting += "NORMALIZE_GS_RATIO__IDF = " + NORMALIZE_GS_RATIO__IDF + "\n";
		setting += "NORMALIZE_GS_RATIO__TEXTLENGTH = " + NORMALIZE_GS_RATIO__TEXTLENGTH + "\n";
		setting += "\n";

		setting += "GET_TOP_LEMMA = " + GET_TOP_LEMMA + "\n";
		setting += "GET_TOP_LEMMA_T = " + GET_TOP_LEMMA_T + "\n";
		setting += "\n";

		setting += "GET_GENRE_SPECIFIC_LEMMA = " + GET_GENRE_SPECIFIC_LEMMA + "\n";
		setting += "GET_GENRE_SPECIFIC_LEMMA_MIN = " + GET_GENRE_SPECIFIC_LEMMA_MIN + "\n";
		setting += "GET_GENRE_SPECIFIC_LEMMA_THRESHOLD = " + GET_GENRE_SPECIFIC_LEMMA_THRESHOLD + "\n";
		setting += "GET_GENRE_SPECIFIC_LEMMA_T = " + GET_GENRE_SPECIFIC_LEMMA_T + "\n";
		setting += "\n";

		setting += "GET_TOP_WORDS = " + GET_TOP_WORDS + "\n";
		setting += "GET_TOP_WORDS_T = " + GET_TOP_WORDS_T + "\n";
		setting += "SKIP_THE_FIRST_X_TOP_WORDS = " + SKIP_THE_FIRST_X_TOP_WORDS + "\n";
		setting += "\n";
		setting += "GET_GENRE_SPECIFIC_WORDS = " + GET_GENRE_SPECIFIC_WORDS + "\n";
		setting += "GET_GENRE_SPECIFIC_WORDS_T = " + GET_GENRE_SPECIFIC_WORDS_T + "\n";
		setting += "GET_GENRE_SPECIFIC_WORDS_MIN = " + GET_GENRE_SPECIFIC_WORDS_MIN + "\n";
		setting += "GET_GENRE_SPECIFIC_WORDS_THRESHOLD = " + GET_GENRE_SPECIFIC_WORDS_THRESHOLD + "\n";
		setting += "\n";
		setting += "GET_TOP_CHARNGRAMS = " + GET_TOP_CHARNGRAMS + "\n";
		setting += "GET_TOP_CHARNGRAMS_T = " + GET_TOP_CHARNGRAMS_T + "\n";
		// setting += "GET_TOP_CHARNGRAMS_MIN = "+ GET_TOP_CHARNGRAMS_MIN + "\n";
		setting += "\n";
		setting += "GET_GENRE_SPECIFIC_CHARNGRAMS = " + GET_GENRE_SPECIFIC_CHARNGRAMS + "\n";
		setting += "GET_GENRE_SPECIFIC_CHARNGRAMS_T = " + GET_GENRE_SPECIFIC_CHARNGRAMS_T + "\n";
		setting += "GET_GENRE_SPECIFIC_CHARNGRAMS_MIN = " + GET_GENRE_SPECIFIC_CHARNGRAMS_MIN + "\n";
		setting += "GET_GENRE_SPECIFIC_CHARNGRAMS_THRESHOLD = " + GET_GENRE_SPECIFIC_CHARNGRAMS_THRESHOLD + "\n";
		setting += "\n";
		setting += "GET_TOP_POSNGRAMS = " + GET_TOP_POSNGRAMS + "\n";
		setting += "GET_TOP_POSNGRAMS_T = " + GET_TOP_POSNGRAMS_T + "\n";
		// setting += "GET_TOP_POSNGRAMS_MIN = " + GET_TOP_POSNGRAMS_MIN + "\n";
		setting += "\n";
		setting += "GET_GENRE_SPECIFIC_POSNGRAMS = " + GET_GENRE_SPECIFIC_POSNGRAMS + "\n";
		setting += "GET_GENRE_SPECIFIC_POSNGRAMS_T = " + GET_GENRE_SPECIFIC_POSNGRAMS_T + "\n";
		setting += "GET_GENRE_SPECIFIC_POSNGRAMS_MIN = " + GET_GENRE_SPECIFIC_POSNGRAMS_MIN + "\n";
		setting += "GET_GENRE_SPECIFIC_POSNGRAMS_THRESHOLD = " + GET_GENRE_SPECIFIC_POSNGRAMS_THRESHOLD + "\n";
		setting += "\n";
		setting += "LOWER_BOUND_FOR_POS_OCCURRENCES = " + LOWER_BOUND_FOR_POS_OCCURRENCES + "\n";
		// setting += "USE_FACTORS_IN_NORMALIZATION = " + USE_FACTORS_IN_NORMALIZATION + "\n";
		// FIXME: write down here what is normalized...
		setting += "\n\n";
		return setting;
	}
}