package de.upb.infexba.ml.feature.ngram;

import java.util.HashMap;

import org.apache.uima.jcas.JCas;

import de.upb.infexba.ml.feature.FeatureHandler;
import de.upb.infexba.ml.feature.GenreFeatureHandler;
import de.upb.infexba.util.CharTools;

public class CharNgramHandler extends NgramHandler {

	static String FEATURE_PREFIX = "charNgram_";

	public CharNgramHandler(int n) {
		super(n);
	}

	public boolean canHandle(String feature) {
		return feature.startsWith(CharNgramHandler.FEATURE_PREFIX);
	}

	String getFeatureName(Object[] ngram) {
		String str = "";
		for (int i = 0; i < ngram.length; i++) {
			Object o = ngram[i];
			if (o instanceof Character) {
				str += getSubstituteChar((Character) o);
			} else
				System.out.println("Error in computeCode of CharNgram: An entry of this.objects is no Character:");
		}
		return CharNgramHandler.FEATURE_PREFIX + str;
	}

	private Character getSubstituteChar(Character c) {
		Character output = '_';
		if (Character.isLetterOrDigit(c))
			output = Character.toLowerCase(c);
		else if (CharTools.isQuotationMark(c))
			output = '`'; // all different kinds of quotation marks will be replaced by "`"
		else if (CharTools.isOpeningBracket(c))
			output = '(';
		else if (CharTools.isClosingBracket(c))
			output = ')';
		else if (CharTools.isBullet(c) || CharTools.isAt(c) || CharTools.isHyphen(c) || CharTools.isCurrencySymbol(c) || CharTools.isTendencySymbol(c)
				|| c.equals("\\") || c.equals("/") || c.equals("&"))
			output = c;
		else if (CharTools.isComma(c) || c.equals(",") || c.equals(";") || c.equals(":") || CharTools.isFullStop(c))
			output = '.'; // symbols like , ; . : ! ? will be replaced by a simple "."
		else {
			output = '_';
		}
		return (Character) output;
	}

	public static String getTextFromFeatureName(String feature) {
		return feature.substring(CharNgramHandler.FEATURE_PREFIX.length() + 1);
	}

	public void handleInMap(JCas jcas, HashMap<String, Double> map) {
		String text = jcas.getDocumentText();
		Character[] window = new Character[this.n];

		// fill first window
		int textIndex = 0;
		int windowIndex = 0;
		while (windowIndex < this.n && textIndex < text.length()) {
			window[windowIndex] = getSubstituteChar(text.charAt(textIndex));

			// ignore "__" (e.g. double whitespace)
			if (windowIndex > 0 && !ignoreChar(window[windowIndex - 1], window[windowIndex]))
				windowIndex++;
			else if (windowIndex == 0)
				windowIndex++;

			textIndex++;
		}
		this.increaseInMap(map, getFeatureName(window), 1.);

		// for (int i = 0; i < window.length; i++)
		// System.out.print(window[i] + " ");
		// System.out.println();

		// handle first window
		while (textIndex < text.length()) {

			// shift left by one
			System.arraycopy(window, 1, window, 0, window.length - 1);

			// next pos
			window[window.length - 1] = getSubstituteChar(text.charAt(textIndex));

			// ignore "__" (e.g. double whitespace)
			if (!ignoreChar(window[window.length - 1], window[window.length - 2]))
				// handle window
				this.increaseInMap(map, getFeatureName(window), 1.);

			// for (int i = 0; i < window.length; i++)
			// System.out.print(window[i] + " ");
			// System.out.println();

			textIndex++;
		}
	}

	private static boolean ignoreChar(char charThatMightBeIgnored, char previousChar) {
		return (previousChar == '_' && charThatMightBeIgnored == '_') || (previousChar == '.' && charThatMightBeIgnored == '.');
	}

	public double getNormalizedValue(String feature, double value, Double numberOfTokens) {
		if (this.normalizationMode == NO_FACTORS)
			return 1. * value / numberOfTokens;

		double factor = 1.;

		if (this.normalizationMode == FeatureHandler.MY_NORMALIZATION) {
			factor = 10;
			if (GenreFeatureHandler.isGenreSpecificFeature(feature))
				factor = 100;
		}
		
		else if(this.normalizationMode == FeatureHandler.CALCULATED_MAXIMA_MUSIC){
			if(feature.equals("charNgram_)_r|gs"))
				factor = 84.0;
			else if(feature.equals("charNgram_***|gs"))
				factor = 5.777777777777778;
			else if(feature.equals("charNgram_+++|gs"))
				factor = 3.8780487804878048;
			else if(feature.equals("charNgram_.di|gs"))
				factor = 43.5;
			else if(feature.equals("charNgram_10_|gs"))
				factor = 12.0;
			else if(feature.equals("charNgram__4_|gs"))
				factor = 41.666666666666664;
			else if(feature.equals("charNgram__5_|gs"))
				factor = 10.869565217391305;
			else if(feature.equals("charNgram__dj|gs"))
				factor = 45.5;
			else if(feature.equals("charNgram__är|gs"))
				factor = 25.25;
			else if(feature.equals("charNgram__â€“_|gs"))
				factor = 43.666666666666664;
			else if(feature.equals("charNgram_agg|gs"))
				factor = 44.0;
			else if(feature.equals("charNgram_arc|gs"))
				factor = 30.666666666666668;
			else if(feature.equals("charNgram_boy|gs"))
				factor = 29.5;
			else if(feature.equals("charNgram_hno|gs"))
				factor = 45.666666666666664;
			else if(feature.equals("charNgram_jam|gs"))
				factor = 38.5;
			else if(feature.equals("charNgram_oy_|gs"))
				factor = 29.5;
			else if(feature.equals("charNgram_oys|gs"))
				factor = 36.0;
			else if(feature.equals("charNgram_p_h|gs"))
				factor = 39.25;
			else if(feature.equals("charNgram_ush|gs"))
				factor = 47.333333333333336;
			else if(feature.equals("charNgram_w_1|gs"))
				factor = 30.0;
			else if(feature.equals("charNgram_w_2|gs"))
				factor = 75.0;
			else if(feature.equals("charNgram_w_4|gs"))
				factor = 80.0;
			else if(feature.equals("charNgram_._d|top"))
				factor = 14.833333333333334;
			else if(feature.equals("charNgram__al|top"))
				factor = 12.5;
			else if(feature.equals("charNgram__an|top"))
				factor = 16.2;
			else if(feature.equals("charNgram__au|top"))
				factor = 13.2;
			else if(feature.equals("charNgram__be|top"))
				factor = 14.181818181818182;
			else if(feature.equals("charNgram__da|top"))
				factor = 10.90909090909091;
			else if(feature.equals("charNgram__de|top"))
				factor = 7.700000000000001;
			else if(feature.equals("charNgram__di|top"))
				factor = 9.916666666666666;
			else if(feature.equals("charNgram__ei|top"))
				factor = 10.375;
			else if(feature.equals("charNgram__er|top"))
				factor = 18.625;
			else if(feature.equals("charNgram__ge|top"))
				factor = 14.624999999999998;
			else if(feature.equals("charNgram__ha|top"))
				factor = 19.75;
			else if(feature.equals("charNgram__in|top"))
				factor = 14.5;
			else if(feature.equals("charNgram__is|top"))
				factor = 17.2;
			else if(feature.equals("charNgram__ma|top"))
				factor = 15.571428571428571;
			else if(feature.equals("charNgram__me|top"))
				factor = 16.875;
			else if(feature.equals("charNgram__mi|top"))
				factor = 14.5;
			else if(feature.equals("charNgram__ni|top"))
				factor = 21.6;
			else if(feature.equals("charNgram__sc|top"))
				factor = 15.400000000000002;
			else if(feature.equals("charNgram__se|top"))
				factor = 15.833333333333332;
			else if(feature.equals("charNgram__si|top"))
				factor = 10.75;
			else if(feature.equals("charNgram__so|top"))
				factor = 18.166666666666664;
			else if(feature.equals("charNgram__st|top"))
				factor = 13.625;
			else if(feature.equals("charNgram__th|top"))
				factor = 10.571428571428571;
			else if(feature.equals("charNgram__un|top"))
				factor = 10.428571428571429;
			else if(feature.equals("charNgram__vo|top"))
				factor = 14.333333333333334;
			else if(feature.equals("charNgram__we|top"))
				factor = 14.5;
			else if(feature.equals("charNgram__wi|top"))
				factor = 17.285714285714285;
			else if(feature.equals("charNgram__zu|top"))
				factor = 11.125;
			else if(feature.equals("charNgram_ach|top"))
				factor = 14.5;
			else if(feature.equals("charNgram_an_|top"))
				factor = 15.1;
			else if(feature.equals("charNgram_and|top"))
				factor = 18.416666666666668;
			else if(feature.equals("charNgram_as_|top"))
				factor = 12.5;
			else if(feature.equals("charNgram_auf|top"))
				factor = 17.8;
			else if(feature.equals("charNgram_ch_|top"))
				factor = 8.826086956521738;
			else if(feature.equals("charNgram_che|top"))
				factor = 9.428571428571429;
			else if(feature.equals("charNgram_cht|top"))
				factor = 10.0;
			else if(feature.equals("charNgram_das|top"))
				factor = 15.625;
			else if(feature.equals("charNgram_den|top"))
				factor = 11.0;
			else if(feature.equals("charNgram_die|top"))
				factor = 9.916666666666666;
			else if(feature.equals("charNgram_e_s|top"))
				factor = 15.0;
			else if(feature.equals("charNgram_eit|top"))
				factor = 14.5;
			else if(feature.equals("charNgram_em_|top"))
				factor = 16.714285714285715;
			else if(feature.equals("charNgram_en.|top"))
				factor = 12.7;
			else if(feature.equals("charNgram_ent|top"))
				factor = 15.0;
			else if(feature.equals("charNgram_er_|top"))
				factor = 6.933333333333334;
			else if(feature.equals("charNgram_ers|top"))
				factor = 16.8;
			else if(feature.equals("charNgram_he_|top"))
				factor = 10.75;
			else if(feature.equals("charNgram_hre|top"))
				factor = 18.0;
			else if(feature.equals("charNgram_ht_|top"))
				factor = 19.25;
			else if(feature.equals("charNgram_ine|top"))
				factor = 13.571428571428573;
			else if(feature.equals("charNgram_ing|top"))
				factor = 18.416666666666668;
			else if(feature.equals("charNgram_lic|top"))
				factor = 13.666666666666668;
			else if(feature.equals("charNgram_lle|top"))
				factor = 18.0;
			else if(feature.equals("charNgram_n._|top"))
				factor = 12.7;
			else if(feature.equals("charNgram_n_a|top"))
				factor = 16.57142857142857;
			else if(feature.equals("charNgram_nd_|top"))
				factor = 11.25;
			else if(feature.equals("charNgram_ng_|top"))
				factor = 15.571428571428571;
			else if(feature.equals("charNgram_ste|top"))
				factor = 14.333333333333334;
		}
		
		else if(this.normalizationMode == FeatureHandler.CALCULATED_MAXIMA_SMARTPHONE){
			 if(feature.equals("charNgram_+++|gs"))
					factor = 9.714285714285715;
				else if(feature.equals("charNgram_++_|gs"))
					factor = 9.714285714285715;
				else if(feature.equals("charNgram_.0.|gs"))
					factor = 27.5;
				else if(feature.equals("charNgram_.0_|gs"))
					factor = 22.0;
				else if(feature.equals("charNgram_.1_|gs"))
					factor = 16.666666666666668;
				else if(feature.equals("charNgram_.__|gs"))
					factor = 31.25;
				else if(feature.equals("charNgram_.co|gs"))
					factor = 28.0;
				else if(feature.equals("charNgram_19_|gs"))
					factor = 29.23943661971831;
				else if(feature.equals("charNgram_2.1|gs"))
					factor = 20.0;
				else if(feature.equals("charNgram___l|gs"))
					factor = 34.0;
				else if(feature.equals("charNgram___r|gs"))
					factor = 17.0;
				else if(feature.equals("charNgram__ht|gs"))
					factor = 17.0;
				else if(feature.equals("charNgram__ic|gs"))
					factor = 13.5;
				else if(feature.equals("charNgram__ir|gs"))
					factor = 41.0;
				else if(feature.equals("charNgram__jo|gs"))
					factor = 22.333333333333336;
				else if(feature.equals("charNgram__mÃ¼|gs"))
					factor = 38.0;
				else if(feature.equals("charNgram__os|gs"))
					factor = 20.0;
				else if(feature.equals("charNgram__ph|gs"))
					factor = 10.214285714285714;
				else if(feature.equals("charNgram__pu|gs"))
					factor = 25.0;
				else if(feature.equals("charNgram__sy|gs"))
					factor = 18.538461538461537;
				else if(feature.equals("charNgram__tw|gs"))
					factor = 8.4;
				else if(feature.equals("charNgram_aar|gs"))
					factor = 50.0;
				else if(feature.equals("charNgram_agt|gs"))
					factor = 25.0;
				else if(feature.equals("charNgram_ckl|gs"))
					factor = 23.75;
				else if(feature.equals("charNgram_dia|gs"))
					factor = 22.0;
				else if(feature.equals("charNgram_dwi|gs"))
					factor = 63.0;
				else if(feature.equals("charNgram_e_2|gs"))
					factor = 18.0;
				else if(feature.equals("charNgram_egi|gs"))
					factor = 40.0;
				else if(feature.equals("charNgram_ets|gs"))
					factor = 19.333333333333332;
				else if(feature.equals("charNgram_hol|gs"))
					factor = 31.0;
				else if(feature.equals("charNgram_htc|gs"))
					factor = 16.0;
				else if(feature.equals("charNgram_i2.|gs"))
					factor = 55.24999999999999;
				else if(feature.equals("charNgram_irg|gs"))
					factor = 55.800000000000004;
				else if(feature.equals("charNgram_izi|gs"))
					factor = 32.0;
				else if(feature.equals("charNgram_med|gs"))
					factor = 30.125;
				else if(feature.equals("charNgram_mei|gs"))
					factor = 24.0;
				else if(feature.equals("charNgram_mmu|gs"))
					factor = 31.5;
				else if(feature.equals("charNgram_mun|gs"))
					factor = 31.5;
				else if(feature.equals("charNgram_mwa|gs"))
					factor = 18.0;
				else if(feature.equals("charNgram_nit|gs"))
					factor = 15.5;
				else if(feature.equals("charNgram_nku|gs"))
					factor = 23.0;
				else if(feature.equals("charNgram_o_e|gs"))
					factor = 50.0;
				else if(feature.equals("charNgram_odc|gs"))
					factor = 11.333333333333332;
				else if(feature.equals("charNgram_ogg|gs"))
					factor = 21.0;
				else if(feature.equals("charNgram_paa|gs"))
					factor = 50.0;
				else if(feature.equals("charNgram_rdp|gs"))
					factor = 18.42857142857143;
				else if(feature.equals("charNgram_reu|gs"))
					factor = 38.5;
				else if(feature.equals("charNgram_ri2|gs"))
					factor = 55.24999999999999;
				else if(feature.equals("charNgram_rn.|gs"))
					factor = 38.0;
				else if(feature.equals("charNgram_sag|gs"))
					factor = 32.5;
				else if(feature.equals("charNgram_set|gs"))
					factor = 27.666666666666668;
				else if(feature.equals("charNgram_sha|gs"))
					factor = 19.333333333333332;
				else if(feature.equals("charNgram_ssc|gs"))
					factor = 17.2;
				else if(feature.equals("charNgram_swe|gs"))
					factor = 34.0;
				else if(feature.equals("charNgram_syn|gs"))
					factor = 14.0;
				else if(feature.equals("charNgram_trÃ¤|gs"))
					factor = 54.0;
				else if(feature.equals("charNgram_tt_|gs"))
					factor = 26.0;
				else if(feature.equals("charNgram_tus|gs"))
					factor = 66.0;
				else if(feature.equals("charNgram_uri|gs"))
					factor = 24.0;
				else if(feature.equals("charNgram_wor|gs"))
					factor = 16.125;
				else if(feature.equals("charNgram_wär|gs"))
					factor = 49.00000000000001;
				else if(feature.equals("charNgram_zah|gs"))
					factor = 13.666666666666668;
				else if(feature.equals("charNgram_._d|top"))
					factor = 14.666666666666668;
				else if(feature.equals("charNgram__an|top"))
					factor = 9.571428571428571;
				else if(feature.equals("charNgram__au|top"))
					factor = 7.25;
				else if(feature.equals("charNgram__be|top"))
					factor = 9.200000000000001;
				else if(feature.equals("charNgram__da|top"))
					factor = 9.0;
				else if(feature.equals("charNgram__de|top"))
					factor = 6.8;
				else if(feature.equals("charNgram__di|top"))
					factor = 6.5;
				else if(feature.equals("charNgram__ei|top"))
					factor = 8.0;
				else if(feature.equals("charNgram__er|top"))
					factor = 13.0;
				else if(feature.equals("charNgram__fÃ¼|top"))
					factor = 12.0;
				else if(feature.equals("charNgram__ge|top"))
					factor = 10.333333333333334;
				else if(feature.equals("charNgram__ha|top"))
					factor = 12.5;
				else if(feature.equals("charNgram__in|top"))
					factor = 9.285714285714285;
				else if(feature.equals("charNgram__ip|top"))
					factor = 5.75;
				else if(feature.equals("charNgram__is|top"))
					factor = 17.0;
				else if(feature.equals("charNgram__ma|top"))
					factor = 9.666666666666666;
				else if(feature.equals("charNgram__mi|top"))
					factor = 10.0;
				else if(feature.equals("charNgram__se|top"))
					factor = 8.5;
				else if(feature.equals("charNgram__si|top"))
					factor = 11.333333333333332;
				else if(feature.equals("charNgram__so|top"))
					factor = 14.666666666666668;
				else if(feature.equals("charNgram__st|top"))
					factor = 15.666666666666668;
				else if(feature.equals("charNgram__un|top"))
					factor = 8.6;
				else if(feature.equals("charNgram__ve|top"))
					factor = 8.5;
				else if(feature.equals("charNgram__vo|top"))
					factor = 10.5;
				else if(feature.equals("charNgram__we|top"))
					factor = 11.230769230769232;
				else if(feature.equals("charNgram__wi|top"))
					factor = 14.0;
				else if(feature.equals("charNgram__zu|top"))
					factor = 11.0;
				else if(feature.equals("charNgram_and|top"))
					factor = 9.0;
				else if(feature.equals("charNgram_as_|top"))
					factor = 9.5;
				else if(feature.equals("charNgram_auf|top"))
					factor = 9.666666666666666;
				else if(feature.equals("charNgram_ch_|top"))
					factor = 5.5;
				else if(feature.equals("charNgram_che|top"))
					factor = 8.857142857142858;
				else if(feature.equals("charNgram_cht|top"))
					factor = 9.666666666666666;
				else if(feature.equals("charNgram_das|top"))
					factor = 9.5;
				else if(feature.equals("charNgram_den|top"))
					factor = 8.333333333333334;
				else if(feature.equals("charNgram_die|top"))
					factor = 6.5;
				else if(feature.equals("charNgram_e_a|top"))
					factor = 11.75;
				else if(feature.equals("charNgram_e_i|top"))
					factor = 13.0;
				else if(feature.equals("charNgram_eit|top"))
					factor = 8.5;
				else if(feature.equals("charNgram_ell|top"))
					factor = 11.0;
				else if(feature.equals("charNgram_em_|top"))
					factor = 10.5;
				else if(feature.equals("charNgram_en.|top"))
					factor = 11.0;
				else if(feature.equals("charNgram_en_|top"))
					factor = 4.5;
				else if(feature.equals("charNgram_ent|top"))
					factor = 12.75;
				else if(feature.equals("charNgram_er_|top"))
					factor = 6.5;
				else if(feature.equals("charNgram_ere|top"))
					factor = 11.333333333333332;
				else if(feature.equals("charNgram_ers|top"))
					factor = 7.75;
				else if(feature.equals("charNgram_fÃ¼r|top"))
					factor = 12.0;
				else if(feature.equals("charNgram_hon|top"))
					factor = 5.666666666666666;
				else if(feature.equals("charNgram_ht_|top"))
					factor = 10.25;
				else if(feature.equals("charNgram_ich|top"))
					factor = 6.846153846153847;
				else if(feature.equals("charNgram_ie_|top"))
					factor = 9.200000000000001;
				else if(feature.equals("charNgram_ine|top"))
					factor = 12.527777777777779;
				else if(feature.equals("charNgram_iph|top"))
					factor = 5.666666666666666;
				else if(feature.equals("charNgram_lic|top"))
					factor = 14.399999999999999;
				else if(feature.equals("charNgram_lle|top"))
					factor = 15.0;
				else if(feature.equals("charNgram_n._|top"))
					factor = 12.833333333333334;
				else if(feature.equals("charNgram_n_s|top"))
					factor = 14.0;
				else if(feature.equals("charNgram_nd_|top"))
					factor = 8.6;
				else if(feature.equals("charNgram_ng_|top"))
					factor = 12.0;
				else if(feature.equals("charNgram_nte|top"))
					factor = 13.6;
				else if(feature.equals("charNgram_pho|top"))
					factor = 5.666666666666666;
				else if(feature.equals("charNgram_ung|top"))
					factor = 9.25;
		}else if (this.normalizationMode == FeatureHandler.CALCULATED_AVERAGE_MUSIC){
			 if(feature.equals("charNgram_)_r|gs"))
					factor = 2043.531316915292;
				else if(feature.equals("charNgram_***|gs"))
					factor = 1841.806374834851;
				else if(feature.equals("charNgram_+++|gs"))
					factor = 2627.3780487804875;
				else if(feature.equals("charNgram_.di|gs"))
					factor = 1210.516072921827;
				else if(feature.equals("charNgram_10_|gs"))
					factor = 1530.5558380280938;
				else if(feature.equals("charNgram__4_|gs"))
					factor = 1173.5176761504488;
				else if(feature.equals("charNgram__5_|gs"))
					factor = 1447.6241889332532;
				else if(feature.equals("charNgram__dj|gs"))
					factor = 1967.5262246165537;
				else if(feature.equals("charNgram__är|gs"))
					factor = 1970.0915979197448;
				else if(feature.equals("charNgram__–_|gs"))
					factor = 2283.067959346696;
				else if(feature.equals("charNgram_agg|gs"))
					factor = 2065.0735715303063;
				else if(feature.equals("charNgram_arc|gs"))
					factor = 929.5051787166672;
				else if(feature.equals("charNgram_boy|gs"))
					factor = 954.1408024705011;
				else if(feature.equals("charNgram_hno|gs"))
					factor = 2054.5114882640933;
				else if(feature.equals("charNgram_jam|gs"))
					factor = 1696.5738502270935;
				else if(feature.equals("charNgram_oy_|gs"))
					factor = 2373.9912039570168;
				else if(feature.equals("charNgram_oys|gs"))
					factor = 2112.097263990505;
				else if(feature.equals("charNgram_p_h|gs"))
					factor = 1529.6767396292175;
				else if(feature.equals("charNgram_ush|gs"))
					factor = 2141.947124565862;
				else if(feature.equals("charNgram_w_1|gs"))
					factor = 683.5278636487204;
				else if(feature.equals("charNgram_w_2|gs"))
					factor = 837.9630424839075;
				else if(feature.equals("charNgram_w_4|gs"))
					factor = 2039.126385074689;
				else if(feature.equals("charNgram_._d|top"))
					factor = 25.29053076484293;
				else if(feature.equals("charNgram__al|top"))
					factor = 31.846242428395957;
				else if(feature.equals("charNgram__an|top"))
					factor = 43.28315288415146;
				else if(feature.equals("charNgram__au|top"))
					factor = 22.51386654982811;
				else if(feature.equals("charNgram__be|top"))
					factor = 27.469952090188368;
				else if(feature.equals("charNgram__da|top"))
					factor = 22.84600362347161;
				else if(feature.equals("charNgram__de|top"))
					factor = 12.526853932394827;
				else if(feature.equals("charNgram__di|top"))
					factor = 17.384967397570353;
				else if(feature.equals("charNgram__ei|top"))
					factor = 19.618220221616895;
				else if(feature.equals("charNgram__er|top"))
					factor = 50.21610062919471;
				else if(feature.equals("charNgram__ge|top"))
					factor = 28.077330278676435;
				else if(feature.equals("charNgram__ha|top"))
					factor = 49.00611938770264;
				else if(feature.equals("charNgram__in|top"))
					factor = 36.369321629334834;
				else if(feature.equals("charNgram__is|top"))
					factor = 55.20575841824511;
				else if(feature.equals("charNgram__ma|top"))
					factor = 38.97381563780572;
				else if(feature.equals("charNgram__me|top"))
					factor = 62.27196426235011;
				else if(feature.equals("charNgram__mi|top"))
					factor = 31.756407464963537;
				else if(feature.equals("charNgram__ni|top"))
					factor = 55.64234205010522;
				else if(feature.equals("charNgram__sc|top"))
					factor = 42.078356170603065;
				else if(feature.equals("charNgram__se|top"))
					factor = 44.21227660386591;
				else if(feature.equals("charNgram__si|top"))
					factor = 30.273370111490106;
				else if(feature.equals("charNgram__so|top"))
					factor = 36.01035051459433;
				else if(feature.equals("charNgram__st|top"))
					factor = 48.64001212677176;
				else if(feature.equals("charNgram__th|top"))
					factor = 86.28983909183296;
				else if(feature.equals("charNgram__un|top"))
					factor = 19.139136532947106;
				else if(feature.equals("charNgram__vo|top"))
					factor = 38.628274689207984;
				else if(feature.equals("charNgram__we|top"))
					factor = 44.76241825649924;
				else if(feature.equals("charNgram__wi|top"))
					factor = 38.86457846356556;
				else if(feature.equals("charNgram__zu|top"))
					factor = 36.089275993648556;
				else if(feature.equals("charNgram_ach|top"))
					factor = 49.28405370640892;
				else if(feature.equals("charNgram_an_|top"))
					factor = 49.63862770954118;
				else if(feature.equals("charNgram_and|top"))
					factor = 50.27451632249576;
				else if(feature.equals("charNgram_as_|top"))
					factor = 35.67839336111958;
				else if(feature.equals("charNgram_auf|top"))
					factor = 43.14501986590043;
				else if(feature.equals("charNgram_ch_|top"))
					factor = 12.684911967587908;
				else if(feature.equals("charNgram_che|top"))
					factor = 24.10967038899895;
				else if(feature.equals("charNgram_cht|top"))
					factor = 27.049209119285884;
				else if(feature.equals("charNgram_das|top"))
					factor = 38.522969202752584;
				else if(feature.equals("charNgram_den|top"))
					factor = 26.65924372599962;
				else if(feature.equals("charNgram_die|top"))
					factor = 17.42354683582396;
				else if(feature.equals("charNgram_e_s|top"))
					factor = 42.69661657609853;
				else if(feature.equals("charNgram_eit|top"))
					factor = 47.320221467092125;
				else if(feature.equals("charNgram_em_|top"))
					factor = 40.77713409088436;
				else if(feature.equals("charNgram_en.|top"))
					factor = 26.48094630478825;
				else if(feature.equals("charNgram_ent|top"))
					factor = 53.843097077335365;
				else if(feature.equals("charNgram_er_|top"))
					factor = 7.9326567011888125;
				else if(feature.equals("charNgram_ers|top"))
					factor = 45.32396429288281;
				else if(feature.equals("charNgram_he_|top"))
					factor = 52.62148627712767;
				else if(feature.equals("charNgram_hre|top"))
					factor = 67.4391786436628;
				else if(feature.equals("charNgram_ht_|top"))
					factor = 41.0208092537798;
				else if(feature.equals("charNgram_ine|top"))
					factor = 25.664645207910343;
				else if(feature.equals("charNgram_ing|top"))
					factor = 55.56732545295227;
				else if(feature.equals("charNgram_lic|top"))
					factor = 39.18003363473114;
				else if(feature.equals("charNgram_lle|top"))
					factor = 50.49782860052791;
				else if(feature.equals("charNgram_n._|top"))
					factor = 23.424613107265532;
				else if(feature.equals("charNgram_n_a|top"))
					factor = 48.45896379602383;
				else if(feature.equals("charNgram_nd_|top"))
					factor = 16.060907489540025;
				else if(feature.equals("charNgram_ng_|top"))
					factor = 54.535753439931156;
				else if(feature.equals("charNgram_ste|top"))
					factor = 36.63957250749763;
			 
		}else if (this.normalizationMode == FeatureHandler.CALCULATED_AVERAGE_SMARTPHONE){
			if(feature.equals("charNgram_+++|gs"))
				factor = 1064.5755950738805;
			else if(feature.equals("charNgram_++_|gs"))
				factor = 1061.719662927202;
			else if(feature.equals("charNgram_.0.|gs"))
				factor = 558.1926679257876;
			else if(feature.equals("charNgram_.0_|gs"))
				factor = 822.1500425547549;
			else if(feature.equals("charNgram_.1_|gs"))
				factor = 428.92890659037533;
			else if(feature.equals("charNgram_.__|gs"))
				factor = 1347.0686667545135;
			else if(feature.equals("charNgram_.co|gs"))
				factor = 649.6995309420139;
			else if(feature.equals("charNgram_19_|gs"))
				factor = 3773.502483750438;
			else if(feature.equals("charNgram_2.1|gs"))
				factor = 619.5722558871736;
			else if(feature.equals("charNgram___l|gs"))
				factor = 633.8718690363902;
			else if(feature.equals("charNgram___r|gs"))
				factor = 782.436816974963;
			else if(feature.equals("charNgram__ht|gs"))
				factor = 353.13978073331157;
			else if(feature.equals("charNgram__ic|gs"))
				factor = 92.21301461331196;
			else if(feature.equals("charNgram__ir|gs"))
				factor = 1091.699364233181;
			else if(feature.equals("charNgram__jo|gs"))
				factor = 522.8214562947138;
			else if(feature.equals("charNgram__mü|gs"))
				factor = 863.5972664639966;
			else if(feature.equals("charNgram__os|gs"))
				factor = 548.273591803844;
			else if(feature.equals("charNgram__ph|gs"))
				factor = 732.5747639955605;
			else if(feature.equals("charNgram__pu|gs"))
				factor = 970.6085314409157;
			else if(feature.equals("charNgram__sy|gs"))
				factor = 659.4901950937357;
			else if(feature.equals("charNgram__tw|gs"))
				factor = 814.5888557849519;
			else if(feature.equals("charNgram_aar|gs"))
				factor = 993.3784937036112;
			else if(feature.equals("charNgram_agt|gs"))
				factor = 833.6411070177421;
			else if(feature.equals("charNgram_ckl|gs"))
				factor = 392.5702992044904;
			else if(feature.equals("charNgram_dia|gs"))
				factor = 971.2835368786284;
			else if(feature.equals("charNgram_dwi|gs"))
				factor = 3505.4156952242506;
			else if(feature.equals("charNgram_e_2|gs"))
				factor = 520.2997587385369;
			else if(feature.equals("charNgram_egi|gs"))
				factor = 1099.3536085605658;
			else if(feature.equals("charNgram_ets|gs"))
				factor = 901.8324176533574;
			else if(feature.equals("charNgram_hol|gs"))
				factor = 990.2586285825457;
			else if(feature.equals("charNgram_htc|gs"))
				factor = 511.6334833349145;
			else if(feature.equals("charNgram_i2.|gs"))
				factor = 5924.875413150169;
			else if(feature.equals("charNgram_irg|gs"))
				factor = 1354.5737381141328;
			else if(feature.equals("charNgram_izi|gs"))
				factor = 540.4271240286574;
			else if(feature.equals("charNgram_med|gs"))
				factor = 923.2484977957504;
			else if(feature.equals("charNgram_mei|gs"))
				factor = 216.4130460113432;
			else if(feature.equals("charNgram_mmu|gs"))
				factor = 1303.6152219755527;
			else if(feature.equals("charNgram_mun|gs"))
				factor = 1005.239787985772;
			else if(feature.equals("charNgram_mwa|gs"))
				factor = 507.31491914351864;
			else if(feature.equals("charNgram_nit|gs"))
				factor = 892.5623883000492;
			else if(feature.equals("charNgram_nku|gs"))
				factor = 735.871581090122;
			else if(feature.equals("charNgram_o_e|gs"))
				factor = 734.3386985849045;
			else if(feature.equals("charNgram_odc|gs"))
				factor = 1158.6280660281695;
			else if(feature.equals("charNgram_ogg|gs"))
				factor = 1001.9531549748358;
			else if(feature.equals("charNgram_paa|gs"))
				factor = 1070.4048295265047;
			else if(feature.equals("charNgram_rdp|gs"))
				factor = 1909.8166571331665;
			else if(feature.equals("charNgram_reu|gs"))
				factor = 681.1106254920086;
			else if(feature.equals("charNgram_ri2|gs"))
				factor = 5814.823427206368;
			else if(feature.equals("charNgram_rn.|gs"))
				factor = 783.8721215193797;
			else if(feature.equals("charNgram_sag|gs"))
				factor = 764.0845061652035;
			else if(feature.equals("charNgram_set|gs"))
				factor = 449.40417362980924;
			else if(feature.equals("charNgram_sha|gs"))
				factor = 615.4886838052144;
			else if(feature.equals("charNgram_ssc|gs"))
				factor = 788.0660230998635;
			else if(feature.equals("charNgram_swe|gs"))
				factor = 777.0409254102207;
			else if(feature.equals("charNgram_syn|gs"))
				factor = 893.3504677841631;
			else if(feature.equals("charNgram_trä|gs"))
				factor = 998.0844740956286;
			else if(feature.equals("charNgram_tt_|gs"))
				factor = 619.7744524090175;
			else if(feature.equals("charNgram_tus|gs"))
				factor = 3187.1975431306137;
			else if(feature.equals("charNgram_uri|gs"))
				factor = 1908.2074681761844;
			else if(feature.equals("charNgram_wor|gs"))
				factor = 290.65318109769765;
			else if(feature.equals("charNgram_wär|gs"))
				factor = 1171.6681747347166;
			else if(feature.equals("charNgram_zah|gs"))
				factor = 643.5131670336021;
			else if(feature.equals("charNgram_._d|top"))
				factor = 28.67765578140487;
			else if(feature.equals("charNgram__an|top"))
				factor = 35.68814702290063;
			else if(feature.equals("charNgram__au|top"))
				factor = 22.12421651796229;
			else if(feature.equals("charNgram__be|top"))
				factor = 24.987822679245102;
			else if(feature.equals("charNgram__da|top"))
				factor = 18.022554489946362;
			else if(feature.equals("charNgram__de|top"))
				factor = 12.431841991655727;
			else if(feature.equals("charNgram__di|top"))
				factor = 19.23228146664678;
			else if(feature.equals("charNgram__ei|top"))
				factor = 19.566557957338578;
			else if(feature.equals("charNgram__er|top"))
				factor = 49.379373421639265;
			else if(feature.equals("charNgram__fü|top"))
				factor = 47.57401043367123;
			else if(feature.equals("charNgram__ge|top"))
				factor = 30.328438364507292;
			else if(feature.equals("charNgram__ha|top"))
				factor = 43.04149076493235;
			else if(feature.equals("charNgram__in|top"))
				factor = 31.611261909580985;
			else if(feature.equals("charNgram__ip|top"))
				factor = 22.924630748799423;
			else if(feature.equals("charNgram__is|top"))
				factor = 61.02052146399811;
			else if(feature.equals("charNgram__ma|top"))
				factor = 41.05940642203539;
			else if(feature.equals("charNgram__mi|top"))
				factor = 34.93457826884171;
			else if(feature.equals("charNgram__se|top"))
				factor = 52.86417834668195;
			else if(feature.equals("charNgram__si|top"))
				factor = 44.47534810514556;
			else if(feature.equals("charNgram__so|top"))
				factor = 47.391657801912224;
			else if(feature.equals("charNgram__st|top"))
				factor = 59.55973316480472;
			else if(feature.equals("charNgram__un|top"))
				factor = 22.22835180201933;
			else if(feature.equals("charNgram__ve|top"))
				factor = 40.66530442762641;
			else if(feature.equals("charNgram__vo|top"))
				factor = 36.25316997541776;
			else if(feature.equals("charNgram__we|top"))
				factor = 38.20380031311575;
			else if(feature.equals("charNgram__wi|top"))
				factor = 38.01833433841522;
			else if(feature.equals("charNgram__zu|top"))
				factor = 37.14481903666983;
			else if(feature.equals("charNgram_and|top"))
				factor = 39.350354738146635;
			else if(feature.equals("charNgram_as_|top"))
				factor = 26.36436945944217;
			else if(feature.equals("charNgram_auf|top"))
				factor = 37.31798399118579;
			else if(feature.equals("charNgram_ch_|top"))
				factor = 13.69327801203469;
			else if(feature.equals("charNgram_che|top"))
				factor = 28.101884926236213;
			else if(feature.equals("charNgram_cht|top"))
				factor = 27.9010189849397;
			else if(feature.equals("charNgram_das|top"))
				factor = 26.46564459698558;
			else if(feature.equals("charNgram_den|top"))
				factor = 26.35011272624405;
			else if(feature.equals("charNgram_die|top"))
				factor = 19.631579448648335;
			else if(feature.equals("charNgram_e_a|top"))
				factor = 45.40559129091164;
			else if(feature.equals("charNgram_e_i|top"))
				factor = 52.28941717491576;
			else if(feature.equals("charNgram_eit|top"))
				factor = 36.02222042302742;
			else if(feature.equals("charNgram_ell|top"))
				factor = 59.33705877120225;
			else if(feature.equals("charNgram_em_|top"))
				factor = 36.67994347940889;
			else if(feature.equals("charNgram_en.|top"))
				factor = 25.818365522116217;
			else if(feature.equals("charNgram_en_|top"))
				factor = 6.878353559714794;
			else if(feature.equals("charNgram_ent|top"))
				factor = 50.027881569145144;
			else if(feature.equals("charNgram_er_|top"))
				factor = 9.022055042687967;
			else if(feature.equals("charNgram_ere|top"))
				factor = 50.01146123636642;
			else if(feature.equals("charNgram_ers|top"))
				factor = 37.646967643407976;
			else if(feature.equals("charNgram_für|top"))
				factor = 46.982639856700104;
			else if(feature.equals("charNgram_hon|top"))
				factor = 20.435533450709066;
			else if(feature.equals("charNgram_ht_|top"))
				factor = 44.621609367493114;
			else if(feature.equals("charNgram_ich|top"))
				factor = 13.398739209693064;
			else if(feature.equals("charNgram_ie_|top"))
				factor = 19.547252230160435;
			else if(feature.equals("charNgram_ine|top"))
				factor = 25.95295000553574;
			else if(feature.equals("charNgram_iph|top"))
				factor = 23.534566290542646;
			else if(feature.equals("charNgram_lic|top"))
				factor = 39.7674670730217;
			else if(feature.equals("charNgram_lle|top"))
				factor = 51.30659481631863;
			else if(feature.equals("charNgram_n._|top"))
				factor = 23.219982871348048;
			else if(feature.equals("charNgram_n_s|top"))
				factor = 46.42579154446238;
			else if(feature.equals("charNgram_nd_|top"))
				factor = 21.72767683606417;
			else if(feature.equals("charNgram_ng_|top"))
				factor = 49.963152108113746;
			else if(feature.equals("charNgram_nte|top"))
				factor = 52.58989396891061;
			else if(feature.equals("charNgram_pho|top"))
				factor = 21.44577304539861;
			else if(feature.equals("charNgram_ung|top"))
				factor = 37.45835500223704;				
			else 
					System.err.println("Fehler bei CharNgram "+feature);
		}
		
		
		return factor * value / numberOfTokens;
	}
}