package de.upb.infexba.ml.feature.vocabulary;

import java.util.Arrays;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.uima.cas.FSIterator;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;

import de.upb.infexba.MonetaryExpression;
import de.upb.infexba.Opinion;
import de.upb.infexba.Sentence;
import de.upb.infexba.TimeExpression;
import de.upb.infexba.ml.feature.FeatureHandler;

/**
 * Handles miscellaneous features.
 * 
 * @author kabu
 * 
 */
public class CounterFeatureHandler extends FeatureHandler {

	// -------------------------------------------------------------------
	// COUNTER-FEATURES
	// -------------------------------------------------------------------

	public static final String NUM_TOKENS = "NUM_TOKENS", SENTENCE_LENGTH = "SENTENCE_LENGTH", NUM_LONG = "NUM_LONG"; // unused
																														// in
																														// COUNTER_ALL

	public static final String WORD_LENGTH = "7_WORD_LENGTH", LIX = "1_LIX";
	public static final String RATIO_OF_CAPTIAL_WORDS = "3_RATIO_OF_CAPTIAL_WORDS", RATIO_OF_ONLY_UPPERCASE_LETTERS = "3_RATIO_OF_ONLY_UPPERCASE_LETTERS",
			RATIO_OF_ONLY_LOWERCASE_LETTERS = "3_RATIO_OF_ONLY_LOWERCASE_LETTERS", RATIO_OF_EXCL_AND_QUESTION_MARKS = "4_RATIO_OF_EXL_AND_QUESTION_MARKS";
	public static final String MONEY_EXPR = "2_MONEY_EXPR", TIME_EXPR = "2_TIME_EXPR";
	public static final String POSITIVE = "6_POSITIVE", NEUTRAL = "6_NEUTRAL", NEGATIVE = "6_NEGATIVE";
	public static final String SMILIES = "5_SMILIES";
	public static String QUOTE = "4_QUOTE", EXCLAMATION_MARK = "4_EXCLAMATION_MARK", DOT = "4_DOT", COLON = "4_COLON", SEMICOLON = "4_SEMICOLON",
			QUESTION_MARK = "4_QUESTION_MARK", PARANTHESIS = "4_PARANTHESIS", KOMMA2 = "4_KOMMA";

	// Some constants

	public static final int NUM_CHARS_OF_LONG_WORD = 7; // number of characters that defines whether a word is long

	public static final int MIN_LENGTH_OF_CAPTIAL_WORD = 5;

	// Lists of this Features

	/**
	 * all except COUNTER_OTHERS
	 */
	public static final String[] COUNTER_ALL = { WORD_LENGTH, LIX, TIME_EXPR, MONEY_EXPR, RATIO_OF_CAPTIAL_WORDS, RATIO_OF_ONLY_UPPERCASE_LETTERS,
			RATIO_OF_ONLY_LOWERCASE_LETTERS, RATIO_OF_EXCL_AND_QUESTION_MARKS, POSITIVE, NEGATIVE, NEUTRAL, SMILIES, QUOTE, EXCLAMATION_MARK, DOT, COLON,
			SEMICOLON, QUESTION_MARK, PARANTHESIS, KOMMA2 };

	public static final String[] COUNTER_WORDSHAPE_AND_PUNCTATION = { RATIO_OF_CAPTIAL_WORDS, RATIO_OF_ONLY_UPPERCASE_LETTERS, RATIO_OF_ONLY_LOWERCASE_LETTERS,
			RATIO_OF_EXCL_AND_QUESTION_MARKS };

	public static final String[] COUNTER_OPTINION = { POSITIVE, NEGATIVE, NEUTRAL };

	public static final String[] COUNTER_EXPR = { TIME_EXPR, MONEY_EXPR };

	public static final String[] COUNTER_OTHERS = { SENTENCE_LENGTH, NUM_LONG, NUM_TOKENS };

	// all that can be handled...
	// TODO: keep up to date

	/**
	 * should contain _all_counter features that can be handled
	 */
	private static final TreeSet<String> all_counter_features;
	static {
		all_counter_features = new TreeSet<String>(Arrays.asList(COUNTER_ALL));
		all_counter_features.addAll(Arrays.asList(COUNTER_OTHERS));
	}

	// -------------------------------------------------------------------
	// Helper.
	// -------------------------------------------------------------------

	Pattern letters = Pattern.compile("\\p{L}+");
	private int smilies = 0;
	private int tokens = 0, sentences = 0;
	private int sum_of_word_lengths = 0;
	private int long_words = 0, captial_words = 0, only_uppercase_letters = 0, only_lowercase_letters = 0;
	private int question_marks = 0, punctation_marks = 0;
	private int quotes = 0, exclamation_marks = 0, dots = 0, kommas = 0, semicolons = 0, colons = 0, paranthesis = 0;
	private int positive = 0, neutral = 0, negative = 0;
	private int time_expressions = 0, money_expressions = 0;

	// -------------------------------------------------------------------
	// FeatureHandler.
	// -------------------------------------------------------------------

	public CounterFeatureHandler() {
		super(COUNTER_ALL);
	}

	public CounterFeatureHandler(String[] counter) {
		super(counter);
	}

	public TreeSet<String> getFeatures() {
		if (this.handles_given_features)
			return this.features;
		else
			return all_counter_features;
	}

	public boolean canHandle(String feature) {
		return all_counter_features.contains(feature);
	}

	// -------------------------------------------------------------------
	// Count.
	// -------------------------------------------------------------------

	public void count_smilies(String text) {
		if (this.handles(SMILIES)) {
			// Compute the smilies feature :)
			Pattern smilie = Pattern.compile("\\s(:-\\)|:\\)|;-\\)|;\\)|:-D|:D|:-\\(|:\\(|;-\\(|;\\(|B-\\)|8-\\)|:-\\*|:-/|:/|=\\))+");
			// Pattern smilie = Pattern.compile("\\s(:|;|=|8)(o|-)?(\\|\\Q)\\E|\\(|/|D|S|P|\\]|\\[|\\}|\\{)+");
			// Pattern smilie = Pattern.compile("\\s[:;=][o-]?[\\Q)(DSP][}{\\/\\E]");
			Matcher matcher = smilie.matcher(text);
			while (matcher.find()) {
				this.smilies++;
				// System.out.println("smilie "+matcher.group());
			}
		}
	}

	private void count_token() {
		this.tokens++;
	}

	private void count_numLongWords(String word) {
		if (word.length() >= NUM_CHARS_OF_LONG_WORD)
			this.long_words++;
	}

	private void count_tokenLength(String word) {
		this.sum_of_word_lengths += word.length();
	}

	private void count_sentence() {
		this.sentences++;
	}

	private void count_timeExpr(JCas jcas) {
		if (this.handles(TIME_EXPR))
			this.time_expressions += jcas.getAnnotationIndex(TimeExpression.type).size();
	}

	private void count_moneyExpr(JCas jcas) {
		if (this.handles(MONEY_EXPR))
			this.money_expressions += jcas.getAnnotationIndex(MonetaryExpression.type).size();
	}

	private void count_capitalWord(String word) {
		if (this.handles(RATIO_OF_CAPTIAL_WORDS)) {
			char first_char = word.charAt(0);
			if (first_char >= 'A' && first_char <= 'Z' && word.length() >= MIN_LENGTH_OF_CAPTIAL_WORD)
				this.captial_words++;
		}
	}

	private void count_lowerCaseWord(String word) {
		if (this.handles(RATIO_OF_ONLY_LOWERCASE_LETTERS)) {
			if (letters.matcher(word).matches() && word.toLowerCase().equals(word)) {
				this.only_lowercase_letters++;
				// System.out.println("this is a lower word: " + word);
			}
		}
	}

	private void count_upperCaseWord(String word) {
		if (this.handles(RATIO_OF_ONLY_UPPERCASE_LETTERS)) {
			if (letters.matcher(word).matches() && word.toUpperCase().equals(word) && word.length() >= MIN_LENGTH_OF_CAPTIAL_WORD) {
				this.only_uppercase_letters++;
				// System.out.println("this is a upper word: " + word);
			}
		}
	}

	private void count_trend(JCas jcas) {
		FSIterator<Annotation> iter = jcas.getAnnotationIndex(Opinion.type).iterator();
		if (iter.hasNext()) {
			String polarity = ((Opinion) iter.next()).getPolarity();
			if (polarity.equalsIgnoreCase("positive"))
				positive++;
			else if (polarity.equalsIgnoreCase("negative"))
				negative++;
			else if (polarity.equalsIgnoreCase("neutral"))
				neutral++;
		}
	}

	private void count_quote(String word) {
		if (VocabularyFeatureHandler.QUOTE_LIST.contains(word))
			quotes++;
	}

	private void count_questionMark(String word) {
		if (word.contains("?")) {
			this.question_marks++;
			this.punctation_marks++;
		}

	}

	private void count_exclamationMark(String word) {
		if (word.contains("!")) {
			exclamation_marks++;
			this.punctation_marks++;
		}
	}

	private void count_dots(String word) {
		if (word.contains(".")) {
			dots++;
			this.punctation_marks++;
		}
	}

	private void count_colon(String word) {
		if (word.contains(":"))
			colons++;
	}

	private void count_semicolons(String word) {
		if (word.contains(";"))
			semicolons++;
	}

	private void count_kommas(String word) {
		if (word.contains(","))
			kommas++;
	}

	private void count_paranthesis(String word) {
		if (VocabularyFeatureHandler.PARANTHESIS_LIST.contains(word))
			paranthesis++;
	}

	// handle-methods

	public void handle(String word) {
		this.count_token();
		this.count_tokenLength(word);
		this.count_numLongWords(word);

		this.count_capitalWord(word);
		this.count_lowerCaseWord(word);
		this.count_upperCaseWord(word);

		this.count_quote(word);
		this.count_colon(word);
		this.count_semicolons(word);
		this.count_exclamationMark(word);
		this.count_questionMark(word);
		this.count_dots(word);
		this.count_kommas(word);
		this.count_paranthesis(word);
	}

	public void handle(Sentence sentence) {
		this.count_sentence();
	}

	public void handle(JCas jcas) {
		this.count_timeExpr(jcas);
		this.count_moneyExpr(jcas);
		this.count_trend(jcas);
		this.count_smilies(jcas.getDocumentText());
	}

	public void computeFeatures() {

		increaseInFeatureMap(NUM_TOKENS, (double) this.tokens);
		increaseInFeatureMap(NUM_LONG, (double) this.long_words / (double) this.tokens);
		increaseInFeatureMap(WORD_LENGTH, ((double) sum_of_word_lengths / (double) this.tokens) / (double) NUM_CHARS_OF_LONG_WORD);
		increaseInFeatureMap(SENTENCE_LENGTH, ((double) this.tokens / (double) this.sentences) / (double) this.tokens);

		increaseInFeatureMap(LIX, ((double) this.tokens / (double) this.sentences + (double) this.long_words / (double) this.tokens * 100.) / 80.);

		increaseInFeatureMap(TIME_EXPR, (double) this.time_expressions / (double) this.sentences);
		increaseInFeatureMap(MONEY_EXPR, (double) this.money_expressions / (double) this.sentences);

		if (this.punctation_marks == 0)
			increaseInFeatureMap(RATIO_OF_EXCL_AND_QUESTION_MARKS, 0.);
		else
			increaseInFeatureMap(RATIO_OF_EXCL_AND_QUESTION_MARKS, (double) (this.exclamation_marks + this.question_marks) / (double) this.punctation_marks);
		increaseInFeatureMap(RATIO_OF_CAPTIAL_WORDS, (double) captial_words / (double) this.tokens);
		increaseInFeatureMap(RATIO_OF_ONLY_LOWERCASE_LETTERS, (double) this.only_lowercase_letters / (double) this.tokens);
		increaseInFeatureMap(RATIO_OF_ONLY_UPPERCASE_LETTERS, (double) this.only_uppercase_letters / (double) this.tokens);

		increaseInFeatureMap(SMILIES, (double) this.smilies);

		increaseInFeatureMap(POSITIVE, (double) positive);
		increaseInFeatureMap(NEGATIVE, (double) negative);
		increaseInFeatureMap(NEUTRAL, (double) neutral);

		increaseInFeatureMap(QUOTE, (double) quotes / (double) this.tokens);
		increaseInFeatureMap(EXCLAMATION_MARK, (double) exclamation_marks / (double) this.tokens);
		increaseInFeatureMap(QUESTION_MARK, (double) question_marks / (double) this.tokens);
		increaseInFeatureMap(COLON, (double) colons / (double) this.tokens);
		increaseInFeatureMap(SEMICOLON, (double) semicolons / (double) this.tokens);
		increaseInFeatureMap(DOT, (double) dots / (double) this.tokens);
		increaseInFeatureMap(KOMMA2, (double) kommas / (double) this.tokens);
		increaseInFeatureMap(PARANTHESIS, (double) paranthesis / (double) this.tokens);
	}

	public void clearFeatureMap() {

		super.clearFeatureMap();

		// these variables are also used during "classification" so clear them as well
		this.smilies = 0;
		this.tokens = 0;
		this.sentences = 0;
		this.sum_of_word_lengths = 0;
		this.long_words = 0;
		this.captial_words = 0;
		this.only_uppercase_letters = 0;
		this.only_lowercase_letters = 0;
		this.question_marks = 0;
		this.punctation_marks = 0;
		this.quotes = 0;
		this.exclamation_marks = 0;
		this.dots = 0;
		this.kommas = 0;
		this.semicolons = 0;
		this.colons = 0;
		this.paranthesis = 0;
		this.positive = 0;
		this.negative = 0;
		this.neutral = 0;
		this.time_expressions = 0;
		this.money_expressions = 0;
	}

	public double getNormalizedValue(String feature, double value, Double noValueNeededHere) {
		if (this.normalizationMode == NO_FACTORS)
			return 1. * value;

		double factor = 1.;

		if (this.normalizationMode == FeatureHandler.MY_NORMALIZATION) {

			if (feature.equals(CounterFeatureHandler.LIX) || feature.equals(CounterFeatureHandler.SENTENCE_LENGTH)
					|| feature.equals(CounterFeatureHandler.WORD_LENGTH))
				factor = 1.;
			else if (feature.equals(CounterFeatureHandler.RATIO_OF_ONLY_UPPERCASE_LETTERS))
				factor = 60.;
			else if (feature.equals(CounterFeatureHandler.RATIO_OF_EXCL_AND_QUESTION_MARKS))
				factor = 2.;
			else if (feature.equals(CounterFeatureHandler.RATIO_OF_CAPTIAL_WORDS))
				factor = 4.;
			else if (feature.equals(CounterFeatureHandler.MONEY_EXPR))
				factor = 20.;
			else if (feature.equals(CounterFeatureHandler.TIME_EXPR))
				factor = 1.;
			else if (feature.equals(CounterFeatureHandler.DOT) || feature.equals(CounterFeatureHandler.QUOTE))
				factor = 50.;
			else if (feature.equals(CounterFeatureHandler.KOMMA2))
				factor = 40.;
			else if (feature.equals(CounterFeatureHandler.EXCLAMATION_MARK) || feature.equals(CounterFeatureHandler.QUESTION_MARK)
					|| feature.equals(CounterFeatureHandler.COLON) || feature.equals(CounterFeatureHandler.SEMICOLON)
					|| feature.equals(CounterFeatureHandler.PARANTHESIS))
				factor = 100.;
		}

		else if (this.normalizationMode == FeatureHandler.CALCULATED_MAXIMA_MUSIC) {
			if(feature.equals("1_LIX"))
				factor = 0.20292136796991167;
			else if(feature.equals("2_MONEY_EXPR"))
				factor = 2.0;
			else if(feature.equals("2_TIME_EXPR"))
				factor = 0.6363636363636364;
			else if(feature.equals("3_RATIO_OF_CAPTIAL_WORDS"))
				factor = 3.1136363636363638;
			else if(feature.equals("3_RATIO_OF_ONLY_LOWERCASE_LETTERS"))
				factor = 1.205607476635514;
			else if(feature.equals("3_RATIO_OF_ONLY_UPPERCASE_LETTERS"))
				factor = 14.333333333333334;
			else if(feature.equals("4_COLON"))
				factor = 14.705882352941176;
			else if(feature.equals("4_DOT"))
				factor = 4.0;
			else if(feature.equals("4_EXCLAMATION_MARK"))
				factor = 8.9;
			else if(feature.equals("4_KOMMA"))
				factor = 8.958333333333334;
			else if(feature.equals("4_PARANTHESIS"))
				factor = 8.555555555555555;
			else if(feature.equals("4_QUESTION_MARK"))
				factor = 13.714285714285714;
			else if(feature.equals("4_QUOTE"))
				factor = 6.71875;
			else if(feature.equals("4_RATIO_OF_EXL_AND_QUESTION_MARKS"))
				factor = 1.0;
			else if(feature.equals("4_SEMICOLON"))
				factor = 21.833333333333332;
			else if(feature.equals("5_SMILIES"))
				factor = 0.5;
			else if(feature.equals("6_NEGATIVE"))
				factor = 1.0;
			else if(feature.equals("6_NEUTRAL"))
				factor = 1.0;
			else if(feature.equals("6_POSITIVE"))
				factor = 1.0;
			else if(feature.equals("7_WORD_LENGTH"))
				factor = 1.0252525252525253;
		}

		else if (this.normalizationMode == FeatureHandler.CALCULATED_MAXIMA_SMARTPHONE) {
			if(feature.equals("1_LIX"))
				factor = 0.3428571428571428;
			else if(feature.equals("2_MONEY_EXPR"))
				factor = 0.6666666666666666;
			else if(feature.equals("2_TIME_EXPR"))
				factor = 0.5;
			else if(feature.equals("3_RATIO_OF_CAPTIAL_WORDS"))
				factor = 2.2698412698412698;
			else if(feature.equals("3_RATIO_OF_ONLY_LOWERCASE_LETTERS"))
				factor = 1.1071428571428572;
			else if(feature.equals("3_RATIO_OF_ONLY_UPPERCASE_LETTERS"))
				factor = 30.0;
			else if(feature.equals("4_COLON"))
				factor = 8.666666666666666;
			else if(feature.equals("4_DOT"))
				factor = 5.5;
			else if(feature.equals("4_EXCLAMATION_MARK"))
				factor = 7.5;
			else if(feature.equals("4_KOMMA"))
				factor = 7.92857142857143;
			else if(feature.equals("4_PARANTHESIS"))
				factor = 3.5749999999999997;
			else if(feature.equals("4_QUESTION_MARK"))
				factor = 10.571428571428571;
			else if(feature.equals("4_QUOTE"))
				factor = 7.75;
			else if(feature.equals("4_RATIO_OF_EXL_AND_QUESTION_MARKS"))
				factor = 1.0;
			else if(feature.equals("4_SEMICOLON"))
				factor = 14.285714285714285;
			else if(feature.equals("5_SMILIES"))
				factor = 0.022727272727272728;
			else if(feature.equals("6_NEGATIVE"))
				factor = 1.0;
			else if(feature.equals("6_NEUTRAL"))
				factor = 1.0;
			else if(feature.equals("6_POSITIVE"))
				factor = 1.0;
			else if(feature.equals("7_WORD_LENGTH"))
				factor = 1.1048088779284833;
			
			else
				System.err.println("Fehler bei Counter "+feature);
			
			
		}else if(this.normalizationMode == FeatureHandler.CALCULATED_AVERAGE_MUSIC){
			if(feature.equals("1_LIX"))
				factor = 0.8546732205456602;
			else if(feature.equals("2_MONEY_EXPR"))
				factor = 72.56760116467585;
			else if(feature.equals("2_TIME_EXPR"))
				factor = 3.938952823982102;
			else if(feature.equals("3_RATIO_OF_CAPTIAL_WORDS"))
				factor = 2.6715285214474465;
			else if(feature.equals("3_RATIO_OF_ONLY_LOWERCASE_LETTERS"))
				factor = 0.9413196422556487;
			else if(feature.equals("3_RATIO_OF_ONLY_UPPERCASE_LETTERS"))
				factor = 260.3898405919373;
			else if(feature.equals("4_COLON"))
				factor = 136.80238843896683;
			else if(feature.equals("4_DOT"))
				factor = 10.720247892954152;
			else if(feature.equals("4_EXCLAMATION_MARK"))
				factor = 124.95605594186618;
			else if(feature.equals("4_KOMMA"))
				factor = 11.249436085535807;
			else if(feature.equals("4_PARANTHESIS"))
				factor = 65.74952502913084;
			else if(feature.equals("4_QUESTION_MARK"))
				factor = 240.14669813144292;
			else if(feature.equals("4_QUOTE"))
				factor = 14.100733803337837;
			else if(feature.equals("4_RATIO_OF_EXL_AND_QUESTION_MARKS"))
				factor = 5.074721692587698;
			else if(feature.equals("4_SEMICOLON"))
				factor = 545.7546215641561;
			else if(feature.equals("5_SMILIES"))
				factor = 39.85294117647059;
			else if(feature.equals("6_NEGATIVE"))
				factor = 7.131578947368421;
			else if(feature.equals("6_NEUTRAL"))
				factor = 2.656862745098039;
			else if(feature.equals("6_POSITIVE"))
				factor = 0.6741293532338309;
			else if(feature.equals("7_WORD_LENGTH"))
				factor = 0.7623415731018216;
		}else if (this.normalizationMode == FeatureHandler.CALCULATED_AVERAGE_SMARTPHONE){
			 if(feature.equals("1_LIX"))
					factor = 0.90688735154998;
				else if(feature.equals("2_MONEY_EXPR"))
					factor = 10.883535860561027;
				else if(feature.equals("2_TIME_EXPR"))
					factor = 4.312936172351535;
				else if(feature.equals("3_RATIO_OF_CAPTIAL_WORDS"))
					factor = 2.412937808768626;
				else if(feature.equals("3_RATIO_OF_ONLY_LOWERCASE_LETTERS"))
					factor = 1.004683935412842;
				else if(feature.equals("3_RATIO_OF_ONLY_UPPERCASE_LETTERS"))
					factor = 935.856858866435;
				else if(feature.equals("4_COLON"))
					factor = 44.5234027661175;
				else if(feature.equals("4_DOT"))
					factor = 9.211977193442058;
				else if(feature.equals("4_EXCLAMATION_MARK"))
					factor = 221.50710515046902;
				else if(feature.equals("4_KOMMA"))
					factor = 14.272732675996236;
				else if(feature.equals("4_PARANTHESIS"))
					factor = 31.66924807220249;
				else if(feature.equals("4_QUESTION_MARK"))
					factor = 176.51879382515364;
				else if(feature.equals("4_QUOTE"))
					factor = 52.215001289114824;
				else if(feature.equals("4_RATIO_OF_EXL_AND_QUESTION_MARKS"))
					factor = 6.298222837350073;
				else if(feature.equals("4_SEMICOLON"))
					factor = 336.47703422237197;
				else if(feature.equals("5_SMILIES"))
					factor = 6.307228915662651;
				else if(feature.equals("6_NEGATIVE"))
					factor = 6.307228915662651;
				else if(feature.equals("6_NEUTRAL"))
					factor = 0.643120393120393;
				else if(feature.equals("6_POSITIVE"))
					factor = 3.49;
				else if(feature.equals("7_WORD_LENGTH"))
					factor = 0.7296076442477277;
		}

		return factor * value;
	}
}
