package de.upb.infexba.util;

import java.util.ArrayList;
import java.util.List;

import org.apache.uima.cas.FSIterator;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;

import de.upb.infexba.Entity;
import de.upb.infexba.RevenueStatement;
import de.upb.infexba.Sentence;
import de.upb.infexba.Token;

/**
 * Utility methods for operations on general annotations like tokens,
 * sentences or entities.
 * 
 * @author hwachsmuth, 2010/10/27
 *
 */
public class AnnoTools {

	// -------------------------------------------------------------------
	// CHECK INDEX
	// -------------------------------------------------------------------

	/**
	 * Returns whether the given entity is new, i.e., it does not overlap
	 * with any of the entity annotations in the index yet
	 * @param jcas The JCas object
	 * @param entity The entity
	 * @return <CODE>true</CODE> iff. the entity is new
	 */
	public static boolean isNewEntity(JCas jcas, Entity entity){
		FSIterator<Annotation> entityIter = jcas.getAnnotationIndex(Entity.type).iterator();
		while (entityIter.hasNext()){
			Entity cur = (Entity) entityIter.next();
			if (overlap(entity, cur))
				return false;
		}
		return true;
	}
	
	
	// -------------------------------------------------------------------
	// COMPARE SPANS
	// -------------------------------------------------------------------

	/**
	 * Returns <CODE>true</CODE> iff. annotation a is left from b.
	 * 
	 * @param a Annotation a
	 * @param b Annotation b
	 * @return <CODE>true</CODE> iff. a is before b
	 */
	public static boolean isBefore(Annotation a, Annotation b){
		if (a.getBegin() < b.getBegin()) return true;
		return false;
	}
	
	/**
	 * Returns <CODE>true</CODE> iff. the text spans represented the two given
	 * annotations do intersect.
	 * @param a Annotation a
	 * @param b Annotation b
	 * @return <CODE>true</CODE> iff. the two annotations intersect
	 */
	public static boolean overlap(Annotation a, Annotation b){
		if (a.getBegin() >= b.getBegin() && a.getBegin() < b.getEnd())
			return true;
		if (a.getEnd() >= b.getBegin() && a.getEnd() <= b.getEnd())
			return true;
		return false;
	}
	
	/**
	 * Returns <CODE>true</CODE> iff. the given annotation a is within the given
	 * annotation b.
	 * 
	 * @param a Annotation a
	 * @param b Annotation b
	 * @return <CODE>true</CODE> iff. a is in b
	 */
	public static boolean isInAnnotation(Annotation a, Annotation b){
		if (a == null || b == null) return false;
		return (a.getBegin() >= b.getBegin() && a.getEnd() <= b.getEnd());
	}
	
	/**
	 * Returns <CODE>true</CODE> iff. the text spans represented by the
	 * given annotations a and b are the same.
	 * 
	 * @param a Annotation a
	 * @param b Annotation b
	 * @return <CODE>true</CODE> iff. the text spans are the same
	 */
	public static boolean isSameTextSpan(Annotation a, Annotation b){
		if (a == null && b == null) return true;
		if (a == null || b == null) return false;
		return a.getBegin() == b.getBegin() && a.getEnd() == b.getEnd();
	}
	
	/**
	 * Returns <CODE>true</CODE> iff. the the two given statements share
	 * the same text span and the same time and money information.
	 * @param a Statement a
	 * @param b Statement b
	 * @return <CODE>true</CODE> iff. the statements are the same
	 */
	public static boolean isSameStatement(RevenueStatement a, RevenueStatement b){
		return (isSameTextSpan(a, b) &&
				a.getTimeExpression().getBegin() == b.getTimeExpression().getBegin() &&
				a.getTimeExpression().getEnd() == b.getTimeExpression().getEnd() &&
				a.getMonetaryExpression().getBegin() == b.getMonetaryExpression().getBegin() &&
				a.getMonetaryExpression().getEnd() == b.getMonetaryExpression().getEnd());
	}
	
	
	// -------------------------------------------------------------------
	// CONVERT
	// -------------------------------------------------------------------

	/**
	 * Converts the given sentence to a list of tokens and returns that list.
	 * 
	 * @param jcas The JCas object
	 * @param sentence The sentence
	 * @return The list of tokens
	 */
	public static List<Token> sentenceToTokens(JCas jcas, Sentence sentence){
		List<Token> tokens = new ArrayList<Token>();
		FSIterator<Annotation> tokenIter = jcas.getAnnotationIndex(Token.type).subiterator(sentence);
		while (tokenIter.hasNext()){
			Token token = (Token) tokenIter.next();
			tokens.add(token);	
		}
		return tokens;
	}
	
	/**
	 * Converts the given statement to a list of tokens and returns that list.
	 * 
	 * @param jcas The JCas object
	 * @param sentence The statement
	 * @return The list of tokens
	 */
	public static List<Token> statementToTokens(JCas jcas, RevenueStatement statement){
		List<Token> tokens = new ArrayList<Token>();
		FSIterator<Annotation> tokenIter = jcas.getAnnotationIndex(Token.type).subiterator(statement);
		while (tokenIter.hasNext()){
			Token token = (Token) tokenIter.next();
			tokens.add(token);	
		}
		return tokens;
	}
	
	/**
	 * Converts the given list of tokens to a list of the strings that refer to
	 * the tokens.
	 * 
	 * @param tokens The list of tokens
	 * @return The list of token strings
	 */
	public static List<String> tokensToStrings(List<Token> tokens){
		List<String> strings = new ArrayList<String>();
		for (Token token : tokens) 
			strings.add(token.getCoveredText());
		return strings;
	}
	
	/**
	 * Converts the given list of tokens to a list of the part-of-speech tags 
	 * that refer to the tokens.
	 * 
	 * @param tokens The list of tokens
	 * @return The list of token strings
	 */
	public static List<String> tokensToPOSTags(List<Token> tokens){
		List<String> strings = new ArrayList<String>();
		for (Token token : tokens) 
			strings.add(token.getPos());
		return strings;
	}
	
	
	// -------------------------------------------------------------------
	// DISTANCES
	// -------------------------------------------------------------------

	/**
	 * Computes the distance in tokens between the two given tokens. Both are always 
	 * supposed to be inside the given annotation.
	 * 
	 * @param jcas The JCas object
	 * @param annotation The annotation
	 * @param token1 The first token
	 * @param token2 The second token
	 * @return The distance measured in tokens
	 */
	public static int getDistanceInTokens(JCas jcas, Annotation annotation, Token token1, Token token2){
		int token1Position = Integer.MIN_VALUE;
		int token2Position = Integer.MAX_VALUE;
		int index = 0;
		FSIterator<Annotation> tokenIter = jcas.getAnnotationIndex(Token.type).subiterator(annotation);
		while (tokenIter.hasNext()){
			Token cur = (Token) tokenIter.next();
			if (cur.equals(token1)) token1Position = index;
			if (cur.equals(token2)) token2Position = index;
			if (token1Position != Integer.MIN_VALUE && token2Position != Integer.MAX_VALUE)
				break;
			index++;
		}
		return (int) Math.abs(token1Position-token2Position);
	}
}
