package de.upb.infexba.util;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * <p>
 * Class that offers methods for accessing or loading the different word 
 * lists used in whole linguistic process, namely 
 * <ul>
 *   <li>a determiner list,</li>
 *   <li>separated prepositions lists for money and time,</li>
 *   <li>a list of connector words,</li>
 *   <li>a fuzzy term list,</li>
 *   <li>a relational term list,</li>
 *   <li>a list of terms that specify boundaries (e.g. "Ende"),</li>
 *   <li>a list of words for mathematic signs,</li>
 *   <li>separated lists for small and big quantity words,</li>
 *   <li>a list of words for ordinal numbers,</li>
 *   <li>separated lists of words for share and multiple fractions,</li>
 *   <li>a currency list,</li>
 *   <li>a list of words for percentage,</li>
 *   <li>an abbreviation list,</li>
 *   <li>a list of sentence begin words</li>
 *   <li>a month list,</li>
 *   <li>a list of words that specify periods,</li>
 *   <li>a list of terms describing revenues</li>
 *   <li>a list of words contradicting the notion of revenue,</li>
 *   <li>...</li>
 * </ul>
 * 
 * The path of the configuration directory has to be set via the according
 * method before any word list can be used.
 * 
 * TODO List in this javadoc description needs to be updated
 * TODO This class is not finished yet!
 * 
 * @author hwachsmuth
 *
 */
public class WordLists {
	
	/**
	 * The path of the configuration directory
	 */
	private static String configurationPath = "conf/";
	
	// -------------------------------------------------------------------
	// PATHS OF WORD LISTS
	// -------------------------------------------------------------------
	
	/** 
	 * The relative path of the word list directory. 
	 */
	private static final String DIRECTORY = "wordlists/";
	

	
	// -------------------------------------------------------------------
	// LOCATIONS OF WORD LISTS
	// -------------------------------------------------------------------
	
	/** The file name of the small word list for abbreviations. */
	public static String ABBREVIATIONS = "abbreviations.txt";
	
	/** The file name of the word list for big quantities. */
	public static String BIG_QUANTITY_WORDS = "money/bigquantitywords.txt";
	
	/** The file name of the word list for words on boundaries of periods. */
	public static String BOUNDARY_WORDS = "time/boundarywords.txt";
	
	/** The file name of the word list for connectors. */
	public static String CONNECTORS = "connectors.txt";

	/** The file name of the word list for time connectors. */
	public static String MONEY_CONNECTORS = "money/connectors.txt";
	
	/** The file name of the big word list for currencies. */
	public static String CURRENCIES = "money/currencies_biglist.txt";
	
	/** The file name of the word list for determiners. */
	public static String DETERMINERS = "determiner.txt";
	
	/** The file name of the word list for fuzzy terms. */
	public static String FUZZY_TERMS = "money/fuzzyterms.txt";
	
	/** The file name of the word list for prepositions of monetary expressions. */
	public static String MONEY_PREPOSITIONS = "money/prepositions_money.txt";
	
	/** The file name of the word list for months. */
	public static String MONTHS = "time/period_months.txt";

	/** The file name of the word list for multiples. */
	public static String MULTIPLE_FRACTIONS = "money/fractions_multiple.txt";

	/** The file name of the word list for ordinal words. */
	public static String ORDINAL_WORDS = "ordinalwords.txt";
	
	/** The file name of the small word list for currencies. */
	public static String PERCENTAGES = "money/percentages.txt";
	
	/** The file name of the word list for time-directional terms. */
	public static String DIRECTIONAL_TERMS = "time/relationalterms.txt";

	/** The file name of the word list for relative periods. */
	public static String RELATIVE_PERIODS = "time/period_rel.txt";
	
	/** The file name of the small word list for sentence begins. */
	public static String SENTENCE_BEGINS = "sentencebegins.txt";
	
	/** The file name of the word list for share. */
	public static String SHARE_FRACTIONS = "money/fractions_share.txt";	
	
	/** The file name of the word list for signs of numbers. */
	public static String SIGN_WORDS = "money/signwords.txt";
	
	/** The file name of the word list for small quantities. */
	public static String SMALL_QUANTITY_WORDS = "money/smallquantitywords.txt";

	/** The file name of the word list for prepositions of time expressions. */
	public static String TIME_PREPOSITIONS = "time/prepositions_time.txt";
	
	/** The file name of the word list for URL beginning substrings. */
	public static String URL_BEGINS = "urlbegins.txt";	
	
	
	
	// -------------------------------------------------------------------
	// LOCATIONS OF FEATURE WORD LISTS
	// -------------------------------------------------------------------
	
	/** The file name of the word list for the property revenue. */
	public static String REVENUE_TERMS = "features/revenue_terms.txt";
	
	/** The file name of the word list for properties different from revenue. */
	public static String OTHER_PROPERTY_TERMS = "features/other_criteria.txt";	
	
	/** The file name of the word list for the property revenue. */
	public static String REVENUE_INDICATORS = "features/revenue_indicators.txt";
	
	/** The file name of the word list for the property revenue. */
	public static String OTHER_INDICATORS = "features/other_indicators.txt";
	
	/** The file name of the word list for words that appear with revenue and change its meaning. */
	public static String PROPERTIES_TRAPS = "features/revenue_traps.txt";	

	/** The file name of the word list for terms that specifiy a comparism. */
	public static String PRECEDING_YEAR_WORDS = "features/preceding_year.txt";
	
	/** The file name of the word list for conjunctions like but, while etc. */
	public static String WORDCLASS_BUT = "features/but.txt";	
	
	/** The file name of the word list for terms that specifiy a comparison. */
	public static String WORDCLASS_COMPARISON = "features/comparison.txt";	
	
	/** The file name of the word list for terms that specifiy a genitive pronoun. */
	public static String WORDCLASS_GENITIVE = "features/genitive.txt";	

	/** The file name of the word list for terms that may not be a named entity */
	public static String ENTITY_RULEOUT_LIST = "features/entity_ruleout.txt";	
	

	/** The file name of the word list for self announce terms. */
	public static String SELF_ANNOUNCE_TERMS = "features/selfannounce.txt";
	
	
	
	// -------------------------------------------------------------------
	// LOCATIONS OF PART-OF-SPEECH LISTS
	// -------------------------------------------------------------------

	/** The file name of the word list for verb part-of-speech tags. */
	public static String VERB_PART_OF_SPEECH = "pos/allverbs_ger.txt";
	
	
	
	// -------------------------------------------------------------------
	// LOCATIONS OF ENTITY LISTS
	// -------------------------------------------------------------------

	/** The file name of the word list for self announce terms. */
	public static String GEOSCOPE_LEMMATA_LEXICON = "lexica/geoscopes_short.txt";

	/** The file name of the word list for market terms. */
	public static String MARKET_LEXICON = "lexica/entity_markets_ger.txt";
	
	/** The file name of the word list for market suffixes. */
	public static String MARKET_SUFFIX_LEXICON = "lexica/entity_marketsuffixes_ger.txt";
	
	/** The file name of the word list for market succeedors. */
	public static String MARKET_SUCCESSOR_LEXICON = "lexica/entity_marketsuccessors_ger.txt";

	/** The file name of the word list for market terms. */
	public static String ORGANIZATION_LEXICON = "lexica/entity_companies.txt";
	
	/** The file name of the word list for market succeedors. */
	public static String ORGANIZATION_SUCCESSOR_LEXICON = "lexica/entity_companysuccessors.txt";
	
	
	// -------------------------------------------------------------------
	// FILE OPERATIONS
	// -------------------------------------------------------------------
	
	/**
	 * Sets the path of the configuration directory. This method MUST be called
	 * with the appropriate path before any word lists is loaded.
	 * 
	 * @param path The path
	 */
	public static void setConfDirectory(String path){
		configurationPath  = path;

		ABBREV_REGEX = loadWordList(ABBREVIATIONS); 
		SENTENCE_BEGIN_REGEX = loadWordList(SENTENCE_BEGINS, false);
	}
	
	/**
	 * Parameter to define whether lines in word lists are rewritten to
	 * both upper and lower case form when they are loaded. Default is
	 * <CODE>true</CODE>.
	 */
	private static boolean toBothCases = true;
	
	/**
	 * Loads the word list given in the file <CODE>filename</CODE>. When
	 * reading the file, this methods checks and rewrites certain inputs,
	 * namely...
	 * <ul> 
	 *   <li>... it replaces signs which have a special meaning in regular 
	 *     expressions,</li>
	 *   <li>... it replaces spaces by a sequence of one or two whitespaces
	 *     because, in this was, line breaks can be recognized, too,
	 *   <li>... it rewrites lines to both upper case and lower case if
	 *     parameter <CODE>toBothCase</CODE> is set to <CODE>true</CODE>, and
	 *   <li>... it skips empty or blank lines because they can harm regular
	 *     expressions severely.</li>
	 * </ul>
	 * Afterwards, this method combines the words of a word list to a pattern
	 * of the form <CODE>(word1|word2|...|wordn)</CODE>.
	 * 
	 * @param filename The name of file to be loaded without its path.
	 * @return The word list as a pattern
	 */
	public static final String loadWordList(String filename){
		// Create list of words from file
		List<String> words = new ArrayList<String>();
		try {
			BufferedReader in = new BufferedReader(new InputStreamReader(
					new FileInputStream(configurationPath + DIRECTORY + filename), "utf8"));
		    String str;
		    while ((str = in.readLine()) != null) {
		    	// Rewrite lines
		    	str = rewriteLine(str);
		    	if (toBothCases) str = rewriteLineToBothCases(str);
		    	// Add word if the line is not be skipped
		    	if (!skipLine(str)) words.add(str);
		    }
		    in.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		// Create regular expression
		StringBuffer regex = new StringBuffer("(");
		for (String word : words) regex.append(word + "|");
		if (regex.charAt(regex.length()-1) == '|')
			regex.deleteCharAt(regex.length()-1);
		regex.append(")");
//		System.out.println(regex.toString() + "\n");
		return regex.toString();
	}
	
	/**
	 * Rewrite some symbols of the given text line to make it conform to
	 * regular expressions.
	 * 
	 * @param line The line to be rewritten
	 * @return The rewritten line
	 */
	private static String rewriteLine(String line){
		line = line.replace(".", "\\.");
    	line = line.replace(" ", "\\s\\s?");
    	line = line.replace("$", "\\$");
    	line = line.replace("(", "\\(");
    	line = line.replace(")", "\\)");
    	line = line.replace("|", "\\|");
    	return line;
	}
	
	/**
	 * Rewrite letters at line begin to unions of both cases, e.g. FOO is 
	 * rewritten as [Ff]OO and foo foo as [Ff]oo foo.
	 * 
	 * @param line The line to be rewritten
	 * @return The rewritten line
	 */
	private static String rewriteLineToBothCases(String line){
		if (Character.isLetter(line.charAt(0))){
			Character otherCase;
			if (Character.isUpperCase(line.charAt(0)))
				otherCase = Character.toLowerCase(line.charAt(0));
			else otherCase = Character.toUpperCase(line.charAt(0));
			StringBuffer buf = new StringBuffer();
			buf.append("[" + line.charAt(0) + otherCase + "]"); 
			if (line.length()>1) buf.append(line.substring(1));
			return buf.toString();
		}	
		return line;
	}
	
	/**
	 * Skip lines that are of length 0 or that start with a
	 * whitespace or with a comment symbol.
	 * 
	 * @param line The line to be checked
	 * @return <CODE>tru</CODE> if line is to be skipped
	 */
	private static boolean skipLine(String line){
		return line.length() == 0 
				|| Character.isWhitespace(line.charAt(0))
				|| CharTools.isComment(line.charAt(0));
	}
	

	/**
	 * Loads the word list given in the file <CODE>filename</CODE> as defined for
	 * the basic <CODE>loadWordList</CODE> method.
	 * <br>
	 * Additionally, the parameter <CODE>bothCases</CODE> sets whether lines in 
	 * word lists are rewritten to both upper and lower case form when they are 
	 * loaded. Default is <CODE>true</CODE>.
	 * <br>
	 * Notice that this parameter is reset after loading the word list.
	 * 
	 * @param filename The name of file to be loaded without its path.
	 * @param bothCases The parameter for upper and lower cases
	 * @return The word list as a pattern
	 */
	public static String loadWordList(String filename, boolean bothCases){
		toBothCases = bothCases;
		String pattern = loadWordList(filename);
		toBothCases = !bothCases;
		return pattern;
	}
	
	/**
	 * Loads the lexicon in the file given by <CODE>filename</CODE>, stores
	 * all its words in a set and returns this set.
	 * 
	 * @param filename The name of the file
	 * @return The lexicon as a set
	 */
	public static Set<String> loadLexicon(String filename){
		Set<String> lexicon = new HashSet<String>();
		try {
		    //BufferedReader in = new BufferedReader(new FileReader(filename));
			BufferedReader in = new BufferedReader(new InputStreamReader(
					new FileInputStream(configurationPath + DIRECTORY + filename), "utf8"));
		    String word;
		    while ((word = in.readLine()) != null){
		    	if (!skipLine(word))
		    		lexicon.add(word);
		    }
		    in.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return lexicon;
	}
	
	
	
	// -------------------------------------------------------------------
	// ACRONYMS
	// -------------------------------------------------------------------
	
	/**
	 * Regular expression for known abbreviations
	 */
	private static String ABBREV_REGEX;
	
	/**
	 * Determines whether the string <code>term</code> is on the
	 * acronym list. Notice that this method is not case sensitve.
	 * 
	 * @param term The term to be looked for
	 * @return <code>true</code iff. the term is on the acronym list.
	 */
	public static boolean isAcronym(String term){
		if (ABBREV_REGEX == null) ABBREV_REGEX = loadWordList(ABBREVIATIONS);
		Matcher matcher = Pattern.compile("\\b"+ABBREV_REGEX.toLowerCase()).matcher(term.toLowerCase());
		if (matcher.find())
			return true;
//		for (Acronym acronym : Acronym.values()) {
//			if (term.toLowerCase().equals(acronym.getAcronym().toLowerCase()))
//				return true;
//		}
		return false;
	}
	
	/**
	 * Determines whether the string <code>term</code> is on the
	 * acronym list. This method is case sensitve.
	 * 
	 * 
	 * @param term The term to be looked for
	 * @return <code>true</code iff. the term is on the acronym list.
	 */	
	public static boolean isExactAcronym(String term){
		for (Acronym acronym : Acronym.values()) {
			if (term.equals(acronym))
				return true;
		}
		return false;
	}
	
	
	
	
	// -------------------------------------------------------------------
	// SENTENCE BEGINS
	// -------------------------------------------------------------------
	
	/**
	 * Regular expression for known abbreviations
	 */
	private static String SENTENCE_BEGIN_REGEX = null;
	
	/**
	 * Determines whether the string <code>word</code> is on the
	 * sentence begin list.
	 * 
	 * @param word The word to be looked for
	 * @return <code>true</code iff. the term is on the acronym list.
	 */
	public static boolean isSentenceBegin(String word){
		if (SENTENCE_BEGIN_REGEX == null) SENTENCE_BEGIN_REGEX = loadWordList(SENTENCE_BEGINS, false);
		Matcher matcher = Pattern.compile(SENTENCE_BEGIN_REGEX).matcher(word);
		if (matcher.find() && matcher.group().length() == word.length())
			return true;
		return false;
	}

	
	
}